from pathlib import Path
from unittest import mock

import pytest
from pytest import param

import shelmet as sh

from .utils import patch_os_fsync


parametrize = pytest.mark.parametrize


def test_dirsync(tmp_path: Path):
    path = tmp_path / "test"
    path.mkdir()

    with patch_os_fsync() as mocked_os_fsync:
        sh.dirsync(path)

    assert mocked_os_fsync.called


def test_fsync__syncs_on_file_object(tmp_path: Path):
    file = tmp_path / "test.txt"

    with file.open("w") as fp:
        fp.write("test")
        fileno = fp.fileno()
        with mock.patch.object(fp, "flush") as mock_flush, patch_os_fsync() as mock_os_fsync:
            sh.fsync(fp)

    assert mock_flush.called
    assert mock_os_fsync.called
    assert mock_os_fsync.call_args[0][0] == fileno


def test_fsync__syncs_on_fileno(tmp_path: Path):
    file = tmp_path / "test.txt"
    file.write_text("test")

    with file.open() as fp:
        fileno = fp.fileno()
        with patch_os_fsync() as mock_os_fsync:
            sh.fsync(fileno)

    assert mock_os_fsync.called
    assert mock_os_fsync.call_args[0][0] == fileno


@parametrize(
    "arg",
    [
        param(1.1),
        param(True),
        param([]),
        param({}),
        param(set()),
    ],
)
def test_fsync__raises_on_invalid_arg_type(arg):
    with pytest.raises(ValueError):
        sh.fsync(arg)



from contextlib import contextmanager
import filecmp
from pathlib import Path
import tarfile
import typing as t
from unittest import mock
import zipfile

import shelmet as sh


try:
    import fcntl
except ImportError:  # pragma: no cover
    fcntl = None  # type: ignore


USES_FCNTL_FULLSYNC = hasattr(fcntl, "F_FULLFSYNC")

TAR_COMPRESSIONS = {
    ".tar": "",
    ".tar.gz": "gz",
    ".tgz": "gz",
    ".taz": "gz",
    ".tar.bz2": "bz2",
    ".tb2": "bz2",
    ".tbz": "bz2",
    ".tbz2": "bz2",
    ".tz2": "bz2",
    ".tar.xz": "xz",
    ".txz": "xz",
}
TAR_EXTENSIONS = list(TAR_COMPRESSIONS.keys())
ZIP_EXTENSIONS = [
    ".docx",
    ".egg",
    ".jar",
    ".odg",
    ".odp",
    ".ods",
    ".odt",
    ".pptx",
    ".xlsx",
    ".zip",
]
ARCHIVE_EXTENSIONS = TAR_EXTENSIONS + ZIP_EXTENSIONS


class File:
    def __init__(self, path: t.Union[Path, str], text: t.Optional[str] = None, size: int = 0):
        self.path = Path(path)
        self.text = text
        self.size = size

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(path={self.path!r})"

    def clone(self) -> "File":
        return self.__class__(self.path, text=self.text, size=self.size)

    def write(self) -> None:
        self.path.parent.mkdir(parents=True, exist_ok=True)

        if self.text is not None:
            self.path.write_text(self.text)
        elif self.size > 0:
            with self.path.open("wb") as fp:
                fp.seek(self.size - 1)
                fp.write(b"\0")
        else:
            self.path.touch()


class Dir:
    def __init__(self, path: t.Union[Path, str], *items: t.Union[File, "Dir"]):
        self.path = Path(path)
        self.items = list(items)

    @property
    def files(self) -> t.List[File]:
        return [item for item in self.items if isinstance(item, File)]

    @property
    def dirs(self) -> t.List["Dir"]:
        return [item for item in self.items if isinstance(item, Dir)]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(path={self.path!r}, files={self.files}, dirs={self.dirs})"
        )

    def clone(self) -> "Dir":
        return self.__class__(self.path, *(item.clone() for item in self.items))

    def mkdir(self) -> None:
        self.path.mkdir(parents=True, exist_ok=True)

        for dir in self.dirs:
            dir.path = self.path / dir.path
            dir.mkdir()

        for file in self.files:
            file.path = self.path / file.path
            file.write()

    def repath(self, root: Path) -> "Dir":
        items = []
        for item in self.items:
            new_path = root / item.path.relative_to(self.path)
            if isinstance(item, File):
                item = item.clone()
                item.path = new_path
            else:
                item = item.repath(new_path)
            items.append(item)
        return Dir(root, *items)


@contextmanager
def patch_os_fsync() -> t.Iterator[mock.MagicMock]:
    if USES_FCNTL_FULLSYNC:
        patched_os_fsync = mock.patch("fcntl.fcntl")
    else:
        patched_os_fsync = mock.patch("os.fsync")

    with patched_os_fsync as mocked_os_fsync:
        yield mocked_os_fsync


def is_same_file(file1: Path, file2: Path) -> bool:
    return filecmp.cmp(file1, file2)


def is_same_dir(dir1: Path, dir2: Path) -> bool:
    return _is_same_dir(filecmp.dircmp(dir1, dir2))


def _is_same_dir(dcmp: filecmp.dircmp) -> bool:
    if dcmp.diff_files or dcmp.left_only or dcmp.right_only:
        return False

    for sub_dcmp in dcmp.subdirs.values():
        if not _is_same_dir(sub_dcmp):
            return False

    return True


def is_subdict(subset: dict, superset: dict) -> bool:
    """Return whether one dict is a subset of another."""
    if isinstance(subset, dict):
        return all(
            key in superset and is_subdict(val, superset[key]) for key, val in subset.items()
        )

    if isinstance(subset, list) and isinstance(superset, list) and len(superset) == len(subset):
        return all(is_subdict(subitem, superset[idx]) for idx, subitem in enumerate(subset))

    # Assume that subset is a plain value if none of the above match.
    return subset == superset


def extract_archive(archive_file: Path, dst: Path, ext: str = "") -> None:
    if not ext:
        ext = next(
            (
                e
                for e in list(TAR_EXTENSIONS) + list(ZIP_EXTENSIONS)
                if archive_file.name.endswith(e)
            ),
            "",
        )

    if ext in TAR_EXTENSIONS:
        _extract_tar(archive_file, dst, ext=ext)
    elif ext in ZIP_EXTENSIONS:
        _extract_zip(archive_file, dst)
    else:
        raise ValueError(f"unrecognized archive extension: {ext}")


def create_archive(archive_file: Path, *sources: Path, ext: str = "") -> None:
    if not ext:
        ext = "".join(archive_file.suffixes)

    if ext in TAR_EXTENSIONS:
        _create_tar(archive_file, *sources, ext=ext)
    elif ext in ZIP_EXTENSIONS:
        _create_zip(archive_file, *sources)
    else:
        raise ValueError(f"unrecognized archive extension: {ext}")


def create_unsafe_archive(archive_file: Path, *sources: Path) -> None:
    ext = "".join(archive_file.suffixes)

    if ext in TAR_EXTENSIONS:
        _create_unsafe_tar(archive_file, *sources)
    elif ext in ZIP_EXTENSIONS:
        _create_unsafe_zip(archive_file, *sources)
    else:
        raise ValueError(f"unrecognized archive extension: {ext}")


def _extract_tar(archive_file: Path, dst: Path, ext: str = "") -> None:
    if not ext:
        ext = "".join(archive_file.suffixes)
    compression = TAR_COMPRESSIONS[ext]
    mode = f"r:{compression}"
    with tarfile.open(archive_file, mode, format=tarfile.PAX_FORMAT) as arc:
        arc.extractall(dst)


def _extract_zip(archive_file: Path, dst: Path) -> None:
    with zipfile.ZipFile(archive_file) as arc:
        arc.extractall(dst)


def _create_tar(archive_file: Path, *sources: Path, ext: str = "") -> None:
    if not ext:
        ext = "".join(archive_file.suffixes)
    compression = TAR_COMPRESSIONS[ext]
    mode = f"w:{compression}"
    with tarfile.open(archive_file, mode, format=tarfile.PAX_FORMAT) as archive:
        for src in sources:
            archive.add(src, arcname=src.name)


def _create_zip(archive_file: Path, *sources: Path) -> None:
    with zipfile.ZipFile(archive_file, "w") as archive:
        for src in sources:
            with sh.cd(src.parent):
                items = [src.relative_to(src.parent), *sh.walk(src.name)]
                for item in items:
                    archive.write(item)


def _create_unsafe_tar(archive_file: Path, src: Path, parent_path: Path) -> None:
    ext = "".join(archive_file.suffixes)
    compression = TAR_COMPRESSIONS[ext]
    mode = f"w:{compression}"
    with tarfile.open(archive_file, mode, format=tarfile.PAX_FORMAT) as archive:
        with sh.cd(src.parent):
            items = [src.relative_to(src.parent)] + list(sh.walk(src.name))
            for item in items:
                member = archive.gettarinfo(str(item))
                member.name = str(parent_path / member.name)
                archive.addfile(member)


def _create_unsafe_zip(archive_file: Path, src: Path, parent_path: Path) -> None:
    with zipfile.ZipFile(archive_file, "w") as archive:
        with sh.cd(src.parent):
            items = [src.relative_to(src.parent)] + list(sh.walk(src.name))
            for item in items:
                member = zipfile.ZipInfo.from_file(str(item))
                member.filename = str(parent_path / member.filename)
                data = item.read_text() if item.is_file() else ""
                archive.writestr(member, data=data)


def create_archive_source(tmp_path: Path, *sources: t.Union[Dir, File]) -> Dir:
    sources = tuple(source.clone() for source in sources)
    src_dir = Dir(tmp_path / "src", *sources)
    src_dir.mkdir()
    return src_dir



from datetime import datetime, timezone
from pathlib import Path
import re
import typing as t

import pytest
from pytest import param

import shelmet as sh

from .utils import ARCHIVE_EXTENSIONS, Dir, File, is_same_dir, is_same_file


parametrize = pytest.mark.parametrize


T_WRITE_FILE = t.Callable[[t.Union[str, Path], str], Path]
DEFAULT_TS_PATTERN = r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+"
DEFAULT_TS_FORMAT = "%Y-%m-%dT%H:%M:%S.%f%z"


@pytest.fixture()
def write_file(tmp_path: Path) -> T_WRITE_FILE:
    def _write_file(path, contents):
        file = tmp_path / path
        file.write_text(contents)
        return file

    return _write_file


@pytest.fixture()
def src_file(tmp_path: Path) -> Path:
    src_file = tmp_path / "src_file.txt"
    src_file.write_text("test")
    return src_file


@pytest.fixture(params=ARCHIVE_EXTENSIONS)
def arc_ext(request) -> str:
    return request.param


def test_backup__backs_up_file(src_file: Path):
    backup_file = sh.backup(src_file)
    assert backup_file != src_file
    assert backup_file.parent == src_file.parent
    assert is_same_file(src_file, backup_file)


def test_backup__backs_up_directory(tmp_path: Path):
    src_dir = Dir(
        tmp_path / "src",
        File("1.txt", text="1"),
        File("2.txt", text="2"),
        File("a/a1.txt", text="a1"),
        File("a/a2.txt", text="a2"),
    )
    src_dir.mkdir()

    backup_dir = sh.backup(src_dir.path)
    assert backup_dir.is_dir()
    assert is_same_dir(src_dir.path, backup_dir)


def test_backup__backs_up_directory_as_archive(tmp_path: Path, arc_ext: str):
    src_dir = Dir(
        tmp_path / "src",
        File("1.txt", text="1"),
        File("2.txt", text="2"),
        File("a/a1.txt", text="a1"),
        File("a/a2.txt", text="a2"),
    )
    src_dir.mkdir()

    backup_archive = sh.backup(src_dir.path, ext=arc_ext)
    assert backup_archive.is_file()

    dst_path = tmp_path / "dst"
    sh.unarchive(backup_archive, dst_path)
    assert is_same_dir(src_dir.path, dst_path / "src")


def test_backup__customizes_backup_parent_directory(tmp_path: Path, src_file: Path):
    dir = tmp_path / "a" / "b" / "c"
    dir.mkdir(parents=True)
    backup_file = sh.backup(src_file, dir=dir)

    assert backup_file.parent == dir
    assert is_same_file(src_file, backup_file)


def test_backup__can_overwrite_destination(src_file: Path):
    dst = src_file.parent / f"{src_file.name}~"
    dst.touch()

    backup_file = sh.backup(src_file, suffix="~", timestamp=None, overwrite=True)
    assert backup_file == dst
    assert is_same_file(src_file, backup_file)


@pytest.mark.freeze_time
def test_backup__appends_local_timestamp_using_strftime(src_file: Path):
    now = datetime.now()
    expected_name = f"{src_file.name}.{now.strftime(DEFAULT_TS_FORMAT)}~"

    backup_file = sh.backup(src_file)
    assert backup_file.name == expected_name


@pytest.mark.freeze_time
def test_backup__appends_utc_timestamp_using_strftime(freezer, src_file: Path):
    utcnow = datetime.now(timezone.utc)
    expected_name = f"{src_file.name}.{utcnow.strftime(DEFAULT_TS_FORMAT)}~"

    backup_file = sh.backup(src_file, utc=True)
    assert backup_file.name == expected_name


@parametrize(
    "filename, args, pattern",
    [
        param("test.txt", {}, re.compile(rf"test\.txt\.{DEFAULT_TS_PATTERN}~")),
        param("test.txt", {"timestamp": "%Y-%d-%m"}, re.compile(r"test\.txt\.\d{4}-\d{2}-\d{2}~")),
        param("test.txt", {"epoch": True}, re.compile(r"test\.txt\.\d+\.\d+~")),
        param("test.txt", {"timestamp": None}, re.compile(r"test\.txt~")),
        param(
            "test.txt", {"prefix": "bak."}, re.compile(rf"bak\.test\.txt\.{DEFAULT_TS_PATTERN}~")
        ),
        param("test.txt", {"suffix": ".bak"}, re.compile(rf"test\.txt\.{DEFAULT_TS_PATTERN}.bak")),
        param("test.txt", {"suffix": ""}, re.compile(rf"test\.txt\.{DEFAULT_TS_PATTERN}")),
        param("test.txt", {"hidden": True}, re.compile(rf"\.test\.txt\.{DEFAULT_TS_PATTERN}~")),
        param(
            "test.txt",
            {"hidden": True, "prefix": "."},
            re.compile(rf"\.test\.txt\.{DEFAULT_TS_PATTERN}~"),
        ),
        param(
            "test.txt",
            {"hidden": True, "prefix": "BACKUP_", "suffix": ".BAK"},
            re.compile(rf"\.BACKUP_test\.txt\.{DEFAULT_TS_PATTERN}\.BAK"),
        ),
        param(
            "test.txt",
            {"namer": lambda src: src.parent / f"{src.name}.bak"},
            re.compile(r"test\.txt\.bak"),
        ),
    ],
)
def test_backup__customizes_filename(
    write_file: T_WRITE_FILE, filename: str, args: t.Dict[str, t.Any], pattern: t.Pattern
):
    src_file = write_file(filename, "test")
    backup_file = sh.backup(src_file, **args)

    assert pattern.fullmatch(backup_file.name), (
        f"Backup of {src_file.name!r} with name {backup_file.name!r}"
        f" did not match pattern {pattern!r}"
    )


@parametrize(
    "timestamp",
    [
        param(True),
        param(False),
        param(b""),
        param(10),
        param({}),
        param([]),
    ],
)
def test_backup__raises_when_timestamp_is_invalid(src_file: Path, timestamp: t.Any):
    with pytest.raises(ValueError):
        sh.backup(src_file, timestamp=timestamp)


def test_backup__raises_when_destination_is_same_as_source(src_file):
    with pytest.raises(FileExistsError):
        sh.backup(src_file, prefix="", suffix="", timestamp=None)


def test_backup__raises_when_destination_exists(src_file):
    dst = src_file.parent / f"{src_file.name}~"
    dst.touch()
    with pytest.raises(FileExistsError):
        sh.backup(src_file, suffix="~", timestamp=None)



"""
The shelmet package.

A shell power-up for working with the file system and running subprocess commands.
"""

__version__ = "0.6.0"

from .archiving import ArchiveError, UnsafeArchiveError, archive, backup, lsarchive, unarchive
from .command import Command, cmd, run
from .fileio import (
    atomicdir,
    atomicfile,
    read,
    readbytes,
    readchunks,
    readlines,
    readtext,
    write,
    writebytes,
    writelines,
    writetext,
)
from .filesystem import (
    chmod,
    chown,
    cp,
    dirsync,
    environ,
    fsync,
    getdirsize,
    mkdir,
    mv,
    rm,
    rmdir,
    rmfile,
    touch,
    umask,
)
from .path import Ls, cd, cwd, homedir, ls, lsdirs, lsfiles, reljoin, walk, walkdirs, walkfiles



from contextlib import ExitStack
from pathlib import Path
import typing as t
from unittest import mock

import pytest
from pytest import param

import shelmet as sh
from shelmet.archiving import ArchiveSource

from .utils import (
    ARCHIVE_EXTENSIONS,
    Dir,
    File,
    create_archive_source,
    extract_archive,
    is_same_dir,
)


parametrize = pytest.mark.parametrize


def _test_archive(
    tmp_path: Path,
    archive_file: Path,
    *sources: t.Union[Dir, File],
    iteratee: t.Callable[[Path], t.Union[str, Path, sh.Ls]] = lambda p: p,
    ext: str = "",
    skip_extraction: bool = False,
):
    src_dir = create_archive_source(tmp_path, *sources)
    sh.archive(archive_file, *(iteratee(source.path) for source in src_dir.items), ext=ext)
    assert archive_file.is_file()

    if skip_extraction:
        return

    dst_path = tmp_path / "dst"
    if len(src_dir.items) > 1:
        extracted_src_path = dst_path / src_dir.path.name
    else:
        extracted_src_path = dst_path

    extract_archive(archive_file, dst_path, ext=ext)

    assert dst_path.is_dir()
    assert extracted_src_path.is_dir()
    assert is_same_dir(src_dir.path, extracted_src_path)


@pytest.fixture(params=ARCHIVE_EXTENSIONS)
def arc_ext(request) -> str:
    """Fixture that yields all archive extensions."""
    return request.param


@pytest.fixture(params=[".tar", ".zip"])
def rep_ext(request) -> str:
    """Fixture that yields a representative sample of archive extensions."""
    return request.param


@parametrize(
    "sources",
    [
        param([File("1.txt", text="1")]),
        param([Dir("a", Dir("b"), File("1.txt", text="1"), File("2.txt", text="2"))]),
        param(
            [
                Dir(
                    "root",
                    Dir(
                        "a",
                        Dir(
                            "aa",
                            Dir("aaa", File("aaa1.txt", text="aaa1"), Dir("aaaa")),
                            File("aa1.txt", text="aa1"),
                        ),
                        File("a1.txt", text="a1"),
                        File("a2.txt", text="a2"),
                    ),
                    Dir("b"),
                    Dir("c"),
                    Dir("d"),
                    File("1.txt", text="1"),
                    File("2.txt", text="2"),
                    File("3.txt", text="3"),
                )
            ]
        ),
        param(
            [
                Dir(
                    "a",
                    Dir(
                        "aa",
                        Dir("aaa", File("aaa1.txt", text="aaa1"), Dir("aaaa")),
                        File("aa1.txt", text="aa1"),
                    ),
                    File("a1.txt", text="a1"),
                    File("a2.txt", text="a2"),
                ),
                Dir("b"),
                Dir("c"),
                Dir("d"),
                File("1.txt", text="1"),
                File("2.txt", text="2"),
                File("3.txt", text="3"),
            ]
        ),
    ],
)
def test_archive__archives_path_sources(
    tmp_path: Path, arc_ext: str, sources: t.List[t.Union[Dir, File]]
):
    archive_file = tmp_path / f"archive{arc_ext}"
    _test_archive(tmp_path, archive_file, *sources)


@parametrize(
    "sources",
    [
        param(
            [
                Dir(
                    "root",
                    Dir(
                        "a",
                        Dir(
                            "aa",
                            Dir("aaa", File("aaa1.txt", text="aaa1"), Dir("aaaa")),
                            File("aa1.txt", text="aa1"),
                        ),
                        File("a1.txt", text="a1"),
                        File("a2.txt", text="a2"),
                    ),
                    Dir("b"),
                    Dir("c"),
                    Dir("d"),
                    File("1.txt", text="1"),
                    File("2.txt", text="2"),
                    File("3.txt", text="3"),
                ),
            ]
        ),
        param(
            [
                Dir(
                    "a",
                    Dir(
                        "aa",
                        Dir("aaa", File("aaa1.txt", text="aaa1"), Dir("aaaa")),
                        File("aa1.txt", text="aa1"),
                    ),
                    File("a1.txt", text="a1"),
                    File("a2.txt", text="a2"),
                ),
                Dir("b"),
                Dir("c"),
                Dir("d"),
            ]
        ),
    ],
)
def test_archive__archives_ls_sources(tmp_path: Path, arc_ext: str, sources: t.List[Dir]):
    archive_file = tmp_path / f"archive{arc_ext}"
    _test_archive(tmp_path, archive_file, *sources, iteratee=sh.walk)


@parametrize(
    "sources, ls_func, expected_listing",
    [
        param(
            [
                Dir(
                    "root",
                    Dir(
                        "a",
                        Dir(
                            "aa",
                            Dir("aaa", File("aaa1.txt", text="aaa1"), Dir("aaaa")),
                            File("aa1.txt", text="aa1"),
                        ),
                        File("a1.txt", text="a1"),
                        File("a2.txt", text="a2"),
                    ),
                    Dir("b"),
                    Dir("c"),
                    Dir("d"),
                    File("1.txt", text="1"),
                    File("2.txt", text="2"),
                    File("3.txt", text="3"),
                ),
            ],
            sh.ls,
            {
                Path("root"),
                Path("root/a"),
                Path("root/b"),
                Path("root/c"),
                Path("root/d"),
                Path("root/1.txt"),
                Path("root/2.txt"),
                Path("root/3.txt"),
            },
        ),
        param(
            [
                Dir(
                    "root",
                    Dir(
                        "a",
                        Dir(
                            "aa",
                            Dir("aaa", File("aaa1.txt", text="aaa1"), Dir("aaaa")),
                            File("aa1.txt", text="aa1"),
                        ),
                        File("a1.txt", text="a1"),
                        File("a2.txt", text="a2"),
                    ),
                    Dir("b"),
                    Dir("c"),
                    Dir("d"),
                    File("1.txt", text="1"),
                    File("2.txt", text="2"),
                    File("3.txt", text="3"),
                ),
            ],
            sh.walkfiles,
            {
                Path("root"),
                Path("root/1.txt"),
                Path("root/2.txt"),
                Path("root/3.txt"),
                Path("root/a/a1.txt"),
                Path("root/a/a2.txt"),
                Path("root/a/aa/aa1.txt"),
                Path("root/a/aa/aaa/aaa1.txt"),
            },
        ),
    ],
)
def test_archive__archives_filtered_ls_sources(
    tmp_path: Path,
    arc_ext: str,
    sources: t.List[Dir],
    ls_func: t.Callable[[Path], sh.Ls],
    expected_listing: t.Set[Path],
):
    archive_file = tmp_path / f"archive{arc_ext}"
    _test_archive(tmp_path, archive_file, *sources, iteratee=ls_func, skip_extraction=True)

    listing = set(sh.lsarchive(archive_file))
    assert listing == expected_listing


def test_archive__allows_extra_leading_file_extension_suffixes(tmp_path: Path, arc_ext: str):
    source = Dir("a", Dir("b"), File("1.txt", text="1"), File("2.txt", text="2"))
    archive_file = tmp_path / f"archive.foo.bar.baz{arc_ext}"
    _test_archive(tmp_path, archive_file, source)


def test_archive__archives_with_explicit_extension_format(tmp_path: Path, arc_ext: str):
    source = Dir("a", Dir("b"), File("1.txt", text="1"), File("2.txt", text="2"))
    archive_file = tmp_path / "archive"
    _test_archive(tmp_path, archive_file, source, ext=arc_ext)


@parametrize(
    "source, root, expected_listing",
    [
        param(
            Dir("a", File("1.txt", text="1"), File("2.txt", text="2"), File("3.txt", text="3")),
            Path("a"),
            {Path("1.txt"), Path("2.txt"), Path("3.txt")},
        )
    ],
)
def test_archive__uses_custom_root_path_inside_archive(
    tmp_path: Path,
    rep_ext: str,
    source: t.Union[File, Dir],
    root: Path,
    expected_listing: t.Set[Path],
):
    src_dir = create_archive_source(tmp_path, source)
    root = src_dir.path / root

    archive_file = tmp_path / f"archive{rep_ext}"
    sh.archive(archive_file, *(item.path for item in src_dir.items), root=root)
    assert archive_file.is_file()

    listing = set(sh.lsarchive(archive_file))
    assert listing == expected_listing


@parametrize(
    "sources, paths, root, repath, expected_listing",
    [
        param(
            [Dir("a", File("1.txt"), File("2.txt"), File("3.txt"))],
            ["a"],
            None,
            "abc",
            {Path("abc"), Path("abc/1.txt"), Path("abc/2.txt"), Path("abc/3.txt")},
        ),
        param(
            [Dir("a", File("1.txt"), File("2.txt"), File("3.txt"))],
            ["a"],
            None,
            {"a": "abc"},
            {Path("abc"), Path("abc/1.txt"), Path("abc/2.txt"), Path("abc/3.txt")},
        ),
        param(
            [
                Dir(
                    "a",
                    Dir("aa1", Dir("aaa1", File("aaa1.txt")), Dir("aaa2", File("aaa2.txt"))),
                    Dir("aa2"),
                ),
                Dir("b"),
                Dir("c"),
            ],
            ["a", "b", "c"],
            None,
            {"a": "1", Path("b"): "2"},
            {
                Path("1"),
                Path("1/aa1"),
                Path("1/aa1/aaa1"),
                Path("1/aa1/aaa1/aaa1.txt"),
                Path("1/aa1/aaa2"),
                Path("1/aa1/aaa2/aaa2.txt"),
                Path("1/aa2"),
                Path("2"),
                Path("src/c"),
            },
        ),
        param(
            [
                Dir(
                    "a",
                    Dir("aa1", Dir("aaa1", File("aaa1.txt")), Dir("aaa2", File("aaa2.txt"))),
                    Dir("aa2"),
                ),
                Dir("b"),
                Dir("c"),
            ],
            ["a", "b", "c"],
            ".",
            {"a": "1", Path("b"): "2"},
            {
                Path("1"),
                Path("1/aa1"),
                Path("1/aa1/aaa1"),
                Path("1/aa1/aaa1/aaa1.txt"),
                Path("1/aa1/aaa2"),
                Path("1/aa1/aaa2/aaa2.txt"),
                Path("1/aa2"),
                Path("2"),
                Path("c"),
            },
        ),
        param(
            [
                Dir(
                    "a",
                    Dir("aa1", Dir("aaa1", File("aaa1.txt")), Dir("aaa2", File("aaa2.txt"))),
                    Dir("aa2"),
                ),
                Dir("b"),
                Dir("c"),
            ],
            [Path("a"), sh.ls("b"), sh.walk("c")],
            ".",
            {"a": "1", "b": "2", "c": "3"},
            {
                Path("1"),
                Path("1/aa1"),
                Path("1/aa1/aaa1"),
                Path("1/aa1/aaa1/aaa1.txt"),
                Path("1/aa1/aaa2"),
                Path("1/aa1/aaa2/aaa2.txt"),
                Path("1/aa2"),
                Path("2"),
                Path("3"),
            },
        ),
    ],
)
def test_archive__repaths_paths_inside_archive(
    tmp_path: Path,
    rep_ext: str,
    sources: t.List[t.Union[File, Dir]],
    paths: t.List[t.Union[str, Path, sh.Ls]],
    root: t.Optional[Path],
    repath: t.Optional[t.Union[str, dict]],
    expected_listing: t.Set[Path],
):
    src_dir = create_archive_source(tmp_path, *sources)
    archive_file = tmp_path / f"archive{rep_ext}"

    with sh.cd(src_dir.path):
        sh.archive(archive_file, *paths, root=root, repath=repath)

    assert archive_file.is_file()

    listing = set(sh.lsarchive(archive_file))
    assert listing == expected_listing


@parametrize(
    "source, root, expected_listing",
    [
        param(
            Dir("a", File("1.txt", text="1"), File("2.txt", text="2"), File("3.txt", text="3")),
            None,
            {Path("a"), Path("a/1.txt"), Path("a/2.txt"), Path("a/3.txt")},
        ),
        param(
            Dir("a", File("1.txt", text="1"), File("2.txt", text="2"), File("3.txt", text="3")),
            Path("a"),
            {Path("1.txt"), Path("2.txt"), Path("3.txt")},
        ),
    ],
)
def test_archive__archives_relative_paths(
    tmp_path: Path,
    rep_ext: str,
    source: t.Union[File, Dir],
    root: t.Optional[Path],
    expected_listing: t.Set[Path],
):
    src_dir = create_archive_source(tmp_path, source)
    archive_file = tmp_path / f"archive{rep_ext}"

    with sh.cd(src_dir.path):
        items = [item.path.relative_to(src_dir.path) for item in src_dir.items]
        sh.archive(archive_file, *items, root=root)

    assert archive_file.is_file()

    listing = set(sh.lsarchive(archive_file))
    assert listing == expected_listing


def test_archive__raises_when_sources_are_not_subpaths_of_root_path(tmp_path: Path, rep_ext: str):
    archive_file = tmp_path / f"archive{rep_ext}"
    with pytest.raises(ValueError) as exc_info:
        sh.archive(archive_file, tmp_path, root="bad-root")
    assert "paths must be a subpath of the root" in str(exc_info.value)


def test_archive__raises_when_file_extension_not_supported(tmp_path: Path):
    with pytest.raises(NotImplementedError) as exc_info:
        sh.archive(tmp_path / "test.txt")
    assert "format not supported" in str(exc_info.value)


def test_archive__raises_when_add_fails(tmp_path: Path, rep_ext: str):
    src_dir = create_archive_source(tmp_path, File("1.txt", text="1"))

    with ExitStack() as mock_stack:
        mock_stack.enter_context(mock.patch("tarfile.TarFile.add", side_effect=Exception))
        mock_stack.enter_context(mock.patch("zipfile.ZipFile.write", side_effect=Exception))

        with pytest.raises(sh.ArchiveError):
            sh.archive(tmp_path / f"archive{rep_ext}", src_dir.path)


@parametrize(
    "paths, repath, expected_error",
    [
        param(["a"], True, "repath must be a string or dict"),
        param(
            ["a", "b"],
            "abc",
            "repath must be a dict when there is more than one archive source path",
        ),
    ],
)
def test_archive__raises_when_repath_is_bad_type(
    tmp_path: Path, paths: list, repath: t.Any, expected_error: str
):
    with pytest.raises(TypeError) as exc_info:
        sh.archive(tmp_path / "archive.tar", *paths, repath=repath)
    assert expected_error in str(exc_info.value)


@parametrize(
    "source, expected",
    [
        param(ArchiveSource("a"), f"ArchiveSource(source='a', path='{sh.cwd() / 'a'}')"),
        param(ArchiveSource(Path("a")), f"ArchiveSource(source='a', path='{sh.cwd() / 'a'}')"),
        param(
            ArchiveSource(sh.ls("a")),
            f"ArchiveSource(source=Ls(path='a', recursive=False), path='{sh.cwd() / 'a'}')",
        ),
        param(
            ArchiveSource(Path("a").absolute()),
            f"ArchiveSource(source='{sh.cwd() / 'a'}', path='{sh.cwd() / 'a'}')",
        ),
    ],
)
def test_archive_source__has_repr(source, expected):
    assert repr(source) == expected



import os
from pathlib import Path

import shelmet as sh


def test_homedir__returns_user_home_directory():
    home = sh.homedir()
    assert isinstance(home, Path)
    assert str(home) == os.path.expanduser("~")



from pathlib import Path

import pytest

import shelmet as sh

from .utils import Dir, File


def test_cp__raises_when_copying_dir_to_existing_file(tmp_path: Path):
    src_dir = tmp_path / "src"
    src_dir.mkdir()

    dst_file = tmp_path / "dst" / "dst.txt"
    dst_file.parent.mkdir()
    dst_file.touch()

    with pytest.raises(FileExistsError):
        sh.cp(src_dir, dst_file)


def test_cp__copies_file_to_file(tmp_path: Path):
    src_file = File("test.txt", text="test")
    src_dir = Dir(tmp_path / "src", src_file)
    src_dir.mkdir()

    dst_file = tmp_path / "dst" / "target.txt"
    sh.cp(src_file.path, dst_file)

    assert dst_file.is_file()
    assert dst_file.read_text() == src_file.text


def test_cp__copies_file_to_existing_dir(tmp_path: Path):
    src_file = File("test.txt", text="test")
    src_dir = Dir(tmp_path / "src", src_file)
    src_dir.mkdir()

    dst_dir = Dir(tmp_path / "dst")
    dst_dir.mkdir()
    sh.cp(src_file.path, dst_dir.path)

    dst_file = dst_dir.path / src_file.path.name
    assert dst_file.is_file()
    assert dst_file.read_text() == src_file.text


def test_cp__copies_dir_to_new_dir(tmp_path: Path):
    src_dir = Dir(
        tmp_path / "src",
        File("1.txt", text="1"),
        File("2.txt", text="2"),
        File("a/a1.txt", text="a1"),
        File("a/a2.txt", text="a2"),
    )
    src_dir.mkdir()
    dst_path = tmp_path / "dst"

    sh.cp(src_dir.path, dst_path)

    copied_src_files = src_dir.repath(dst_path).files
    for file in copied_src_files:
        assert file.path.is_file()
        assert file.path.read_text() == file.text


def test_cp__copies_and_merge_dir_to_existing_dir(tmp_path: Path):
    src_dir = Dir(
        tmp_path / "src",
        File("1.txt", text="1"),
        File("2.txt", text="2"),
        File("a/a1.txt", text="a1"),
        File("a/a2.txt", text="a2"),
    )
    src_dir.mkdir()

    dst_dir = Dir(
        tmp_path / "dst",
        File("11.txt", text="11"),
        File("22.txt", text="22"),
        File("a/b1.txt", text="b1"),
        File("a/b2.txt", text="b2"),
    )
    dst_dir.mkdir()

    sh.cp(src_dir.path, dst_dir.path)

    copied_src_files = src_dir.repath(dst_dir.path).files
    all_files = copied_src_files + dst_dir.files

    for file in all_files:
        assert file.path.is_file()
        assert file.path.read_text() == file.text



from pathlib import Path
import typing as t

import pytest
from pytest import param

import shelmet as sh

from .utils import Dir, File


parametrize = pytest.mark.parametrize


@parametrize(
    "sources",
    [
        param([File("1.txt")], id="one_file"),
        param([File("1.txt"), File("2.txt"), File("3.txt")], id="many_files"),
        param([Dir("1")], id="one_dir_with_no_files"),
        param([Dir("1", File("1.txt"), File("2.txt"))], id="one_dir_with_files"),
        param(
            [Dir("1"), Dir("2"), Dir("3/4"), Dir("5/6/7")],
            id="many_dirs_with_no_files",
        ),
        param(
            [
                Dir("1", File("1.txt")),
                Dir("2", File("2.txt"), Dir("2.1")),
                Dir("3/4", File("3.txt"), File("4.txt")),
                Dir("5/6/7"),
            ],
            id="many_dirs_with_files",
        ),
    ],
)
def test_rm(tmp_path: Path, sources: t.Sequence[File]):
    base_dir = Dir(tmp_path, *sources)
    base_dir.mkdir()

    for src in sources:
        assert src.path.exists()

    sh.rm(*(src.path for src in sources))

    for src in sources:
        assert not src.path.exists()


@parametrize(
    "sources",
    [
        param([Dir("1")], id="one_dir_with_no_files"),
        param([Dir("1", File("1.txt"), File("2.txt"))], id="one_dir_with_files"),
        param(
            [Dir("1"), Dir("2"), Dir("3/4"), Dir("5/6/7")],
            id="many_dirs_with_no_files",
        ),
        param(
            [
                Dir("1", File("1.txt")),
                Dir("2", File("2.txt"), Dir("2.1")),
                Dir("3/4", File("3.txt"), File("4.txt")),
                Dir("5/6/7"),
            ],
            id="many_dirs_with_files",
        ),
    ],
)
def test_rmdir(tmp_path: Path, sources: t.Sequence[File]):
    base_dir = Dir(tmp_path, *sources)
    base_dir.mkdir()

    for src in sources:
        assert src.path.exists()

    sh.rmdir(*(src.path for src in sources))

    for src in sources:
        assert not src.path.exists()


def test_rmdir__raises_on_file(tmp_path: Path):
    path = tmp_path / "test.txt"
    path.touch()

    with pytest.raises(NotADirectoryError):
        sh.rmdir(path)


@parametrize(
    "sources",
    [
        param([File("1.txt")], id="one_file"),
        param([File("1.txt"), File("2.txt"), File("3.txt")], id="many_files"),
    ],
)
def test_rmfile(tmp_path: Path, sources: t.Sequence[File]):
    base_dir = Dir(tmp_path, *sources)
    base_dir.mkdir()

    for src in sources:
        assert src.path.exists()

    sh.rmfile(*(src.path for src in sources))

    for src in sources:
        assert not src.path.exists()


def test_rmfile__raises_on_dir(tmp_path: Path):
    path = tmp_path / "test"
    path.mkdir()

    with pytest.raises(OSError):
        sh.rmfile(path)


@parametrize(
    "rm_fn",
    [
        param(sh.rm),
        param(sh.rmdir),
        param(sh.rmfile),
    ],
)
def test_rm__ignores_missing_sources(tmp_path: Path, rm_fn: t.Callable):
    rm_fn(tmp_path / "1", tmp_path / "2", tmp_path / "3")



from pathlib import Path
import typing as t

import pytest
from pytest import param

import shelmet as sh

from .utils import ARCHIVE_EXTENSIONS, Dir, File, create_archive, create_archive_source


parametrize = pytest.mark.parametrize


@pytest.fixture(params=ARCHIVE_EXTENSIONS)
def arc_ext(request) -> str:
    """Fixture that yields all archive extensions."""
    return request.param


@parametrize(
    "source, expected",
    [
        param(
            Dir(
                "a",
                Dir("b"),
                Dir("c", File("c1", text="c1"), File("c2", text="c2")),
                Dir("d", Dir("e", Dir("f", File("f1", text="f1")), File("e1", text="e1"))),
                File("a1", text="a1"),
                File("a2", text="a2"),
            ),
            {
                Path("a"),
                Path("a/b"),
                Path("a/c"),
                Path("a/c/c1"),
                Path("a/c/c2"),
                Path("a/d"),
                Path("a/d/e"),
                Path("a/d/e/f"),
                Path("a/d/e/f/f1"),
                Path("a/d/e/e1"),
                Path("a/a1"),
                Path("a/a2"),
            },
        )
    ],
)
def test_lsarchive__returns_list_of_archive_members(
    tmp_path: Path, arc_ext: str, source: Dir, expected: t.Set[Path]
):
    archive_file = tmp_path / f"archive{arc_ext}"
    src_dir = create_archive_source(tmp_path, source)
    create_archive(archive_file, src_dir.items[0].path)

    listing = sh.lsarchive(archive_file)
    assert set(listing) == expected


@parametrize(
    "source, expected",
    [
        param(
            Dir(
                "a",
                Dir("b"),
                Dir("c", File("c1", text="c1"), File("c2", text="c2")),
                Dir("d", Dir("e", Dir("f", File("f1", text="f1")), File("e1", text="e1"))),
                File("a1", text="a1"),
                File("a2", text="a2"),
            ),
            {
                Path("a"),
                Path("a/b"),
                Path("a/c"),
                Path("a/c/c1"),
                Path("a/c/c2"),
                Path("a/d"),
                Path("a/d/e"),
                Path("a/d/e/f"),
                Path("a/d/e/f/f1"),
                Path("a/d/e/e1"),
                Path("a/a1"),
                Path("a/a2"),
            },
        )
    ],
)
def test_lsarchive__returns_list_of_archive_members_with_explicit_extension_format(
    tmp_path: Path, arc_ext: str, source: Dir, expected: t.Set[Path]
):
    archive_file = tmp_path / "archive"
    src_dir = create_archive_source(tmp_path, source)
    create_archive(archive_file, src_dir.items[0].path, ext=arc_ext)

    listing = sh.lsarchive(archive_file, ext=arc_ext)
    assert set(listing) == expected



from pathlib import Path
import typing as t

import pytest
from pytest import param

import shelmet as sh


parametrize = pytest.mark.parametrize


@parametrize(
    "paths",
    [
        param(["a"]),
        param(["a", "b", "c/d/e"]),
    ],
)
def test_touch(tmp_path: Path, paths: t.List[str]):
    targets = [tmp_path / path for path in paths]
    sh.touch(*targets)
    for path in targets:
        assert path.is_file()



from pathlib import Path
import typing as t

import pytest
from pytest import param

import shelmet as sh

from .utils import Dir, File, patch_os_fsync


parametrize = pytest.mark.parametrize


@parametrize(
    "opts",
    [
        param({}),
        param({"overwrite": False}),
        param({"skip_sync": True}),
        param({"overwrite": False, "skip_sync": True}),
    ],
)
def test_atomicdir(tmp_path: Path, opts: t.Dict[str, t.Any]):
    dir_path = tmp_path / "test"
    files = [
        File("1.txt", text="1"),
        File("2.txt", text="2"),
        File("3.txt", text="3"),
        File("a/a.txt", text="a"),
        File("b/b.txt", text="b"),
        File("c/c.txt", text="c"),
    ]

    with sh.atomicdir(dir_path, **opts) as stage_path:
        assert stage_path.exists()
        assert not dir_path.exists()

        stage_dir = Dir(stage_path, *files)
        stage_dir.mkdir()

        assert not dir_path.exists()

    assert dir_path.exists()

    created_files = stage_dir.repath(dir_path).files
    for file in created_files:
        assert file.path.exists()
        assert file.path.read_text() == file.text


def test_atomicdir__syncs_dir(tmp_path: Path):
    dir = tmp_path / "test"

    with patch_os_fsync() as mocked_os_fsync:
        with sh.atomicdir(dir):
            pass

    assert mocked_os_fsync.called
    assert mocked_os_fsync.call_count == 2


def test_atomicdir__skips_sync_when_disabled(tmp_path: Path):
    dir = tmp_path / "test"

    with patch_os_fsync() as mocked_os_fsync:
        with sh.atomicdir(dir, skip_sync=True):
            pass

    assert not mocked_os_fsync.called


def test_atomicdir__overwrites_when_enabled(tmp_path: Path):
    dir = Dir(tmp_path / "test", File("1"), File("2"), File("3"))
    dir.mkdir()

    assert list(dir.path.iterdir())

    with sh.atomicdir(dir.path):
        pass

    assert not list(dir.path.iterdir())


def test_atomicdir__does_not_overwrite_when_disabled(tmp_path: Path):
    dir = Dir(tmp_path / "test", File("1"), File("2"), File("3"))
    dir.mkdir()

    with pytest.raises(FileExistsError):
        with sh.atomicdir(dir.path, overwrite=False):
            pass


def test_atomicdir__fails_if_path_is_file(tmp_path: Path):
    already_exists_file = tmp_path / "test"
    already_exists_file.write_text("")

    with pytest.raises(FileExistsError):
        with sh.atomicdir(already_exists_file):
            pass


@parametrize(
    "opts",
    [
        param({}),
        param({"overwrite": False}),
        param({"skip_sync": True}),
        param({"overwrite": False, "skip_sync": True}),
    ],
)
def test_atomicfile(tmp_path: Path, opts: t.Dict[str, t.Any]):
    file = tmp_path / "test.txt"
    text = "test"

    with sh.atomicfile(file, **opts) as fp:
        assert not file.exists()
        fp.write(text)
        assert not file.exists()

    assert file.exists()
    assert file.read_text() == text


def test_atomicfile__syncs_new_file_and_dir(tmp_path: Path):
    file = tmp_path / "test.txt"

    with patch_os_fsync() as mocked_os_fsync:
        with sh.atomicfile(file) as fp:
            fp.write("test")

    assert mocked_os_fsync.called
    assert mocked_os_fsync.call_count == 2


def test_atomicfile__skips_sync_when_disabled(tmp_path: Path):
    file = tmp_path / "test.txt"

    with patch_os_fsync() as mocked_os_fsync:
        with sh.atomicfile(file, skip_sync=True) as fp:
            fp.write("test")

    assert not mocked_os_fsync.called


def test_atomicfile__does_not_overwrite_when_disabled(tmp_path: Path):
    file = tmp_path / "test.txt"
    file.write_text("")

    with pytest.raises(FileExistsError):
        with sh.atomicfile(file, overwrite=False):
            pass


def test_atomicfile__fails_if_path_is_dir(tmp_path: Path):
    already_exists_dir = tmp_path
    with pytest.raises(IsADirectoryError):
        with sh.atomicfile(already_exists_dir):
            pass

    will_exist_dir = tmp_path / "test"
    with pytest.raises(IsADirectoryError):
        with sh.atomicfile(will_exist_dir) as fp:
            will_exist_dir.mkdir()
            fp.write("test")


@parametrize(
    "mode",
    [
        param("r"),
        param("r+"),
        param("rb"),
        param("rb+"),
        param("a"),
        param("a+"),
        param("ab"),
        param("ab+"),
        param("x"),
        param("x+"),
        param(True),
    ],
)
def test_atomicfile__raises_when_mode_invalid(tmp_path: Path, mode: t.Any):
    with pytest.raises(ValueError):
        with sh.atomicfile(tmp_path / "test.txt", mode):
            pass



import fnmatch
from pathlib import Path
import re
import typing as t
from unittest import mock

import pytest
from pytest import param

import shelmet as sh
from shelmet.types import LsFilter

from .utils import Dir, File


parametrize = pytest.mark.parametrize


@parametrize(
    "items, kwargs, expected_contents",
    [
        param(
            [
                Dir("x/xx", File("x1.txt")),
                Dir("y/yy", File("y1.txt"), File("y2.txt")),
                Dir("z/zz"),
                File("a.txt"),
                File("b.txt"),
                File("c.txt"),
            ],
            {},
            {Path("x"), Path("y"), Path("z"), Path("a.txt"), Path("b.txt"), Path("c.txt")},
        ),
        param(
            [
                Dir("x/xx", File("x1.txt")),
                Dir("y/yy", File("y1.txt"), File("y2.txt")),
                Dir("z/zz"),
                File("a.txt"),
                File("b.txt"),
                File("c.txt"),
            ],
            {"recursive": True},
            {
                Path("x"),
                Path("x/xx"),
                Path("x/xx/x1.txt"),
                Path("y"),
                Path("y/yy"),
                Path("y/yy/y1.txt"),
                Path("y/yy/y2.txt"),
                Path("z"),
                Path("z/zz"),
                Path("a.txt"),
                Path("b.txt"),
                Path("c.txt"),
            },
        ),
        param(
            [
                Dir("x/xx", File("x1.txt")),
                Dir("y/yy", File("y1.txt"), File("y2.txt")),
                Dir("z/zz"),
                File("a.txt"),
                File("b.txt"),
                File("c.txt"),
            ],
            {"recursive": True, "only_files": True},
            {
                Path("x/xx/x1.txt"),
                Path("y/yy/y1.txt"),
                Path("y/yy/y2.txt"),
                Path("a.txt"),
                Path("b.txt"),
                Path("c.txt"),
            },
        ),
        param(
            [
                Dir("x/xx", File("x1.txt")),
                Dir("y/yy", File("y1.txt"), File("y2.txt")),
                Dir("z/zz"),
                File("a.txt"),
                File("b.txt"),
                File("c.txt"),
            ],
            {"recursive": True, "only_dirs": True},
            {Path("x"), Path("x/xx"), Path("y"), Path("y/yy"), Path("z"), Path("z/zz")},
        ),
        param(
            [
                Dir("x/xx", File("x1.txt")),
                Dir("y/yy", File("y1.txt"), File("y2.txt")),
                Dir("z/zz"),
                File("a.txt"),
                File("b.txt"),
                File("c.txt"),
            ],
            {"include": "*.txt"},
            {Path("a.txt"), Path("b.txt"), Path("c.txt")},
        ),
        param(
            [
                Dir("x/xx", File("x1.txt")),
                Dir("y/yy", File("y1.txt"), File("y2.txt")),
                Dir("z/zz"),
                File("a.txt"),
                File("b.txt"),
                File("c.txt"),
            ],
            {"exclude": "*.txt"},
            {Path("x"), Path("y"), Path("z")},
        ),
        param(
            [
                Dir("x/xx", File("x1.txt")),
                Dir("y/yy", File("y1.txt"), File("y2.txt")),
                Dir("z/zz"),
                File("a.txt"),
                File("b.txt"),
                File("c.txt"),
            ],
            {"include": "*.txt", "recursive": True},
            {
                Path("x/xx/x1.txt"),
                Path("y/yy/y1.txt"),
                Path("y/yy/y2.txt"),
                Path("a.txt"),
                Path("b.txt"),
                Path("c.txt"),
            },
        ),
        param(
            [
                Dir("x/xx", File("x1.txt")),
                Dir("y/yy", File("y1.txt"), File("y2.txt")),
                Dir("z/zz"),
                File("a.txt"),
                File("b.txt"),
                File("c.txt"),
            ],
            {"exclude": "*.txt", "recursive": True},
            {
                Path("x"),
                Path("x/xx"),
                Path("y"),
                Path("y/yy"),
                Path("z"),
                Path("z/zz"),
            },
        ),
        param(
            [
                Dir("x/xx", File("x1.txt")),
                Dir("y/yy", File("y1.txt"), File("y2.txt")),
                Dir("z/zz"),
                File("a.txt"),
                File("b.txt"),
                File("c.txt"),
            ],
            {"include": ["*.txt"], "only_dirs": True, "recursive": True},
            set(),
        ),
    ],
)
def test_ls(
    tmp_path: Path,
    items: t.List[t.Union[Dir, File]],
    kwargs: dict,
    expected_contents: t.Set[Path],
):
    src = Dir(tmp_path, *items)
    src.mkdir()
    with sh.cd(tmp_path):
        contents = set(sh.ls("", **kwargs))
    assert contents == expected_contents


@parametrize(
    "include",
    [
        param("*_include", id="str"),
        param(["foo", "*_include"], id="str_list"),
        param(re.compile(fnmatch.translate("*_include")), id="regex"),
        param(
            [re.compile(fnmatch.translate("foo")), re.compile(fnmatch.translate("*_include"))],
            id="regex_list",
        ),
        param(lambda p: p.name.endswith("_include"), id="callable"),
        param(
            [lambda p: p.name == "foo", lambda p: p.name.endswith("_include")], id="callable_list"
        ),
    ],
)
def test_ls__includes_on_multiple_types(tmp_path: Path, include: LsFilter):
    items: t.List[t.Union[Dir, File]] = [
        Dir("a_dir_include"),
        Dir("b_dir"),
        Dir("c_dir_include"),
        File("d_file_include"),
        File("e_file"),
        File("f_file_include"),
    ]
    expected_contents = {
        Path("a_dir_include"),
        Path("c_dir_include"),
        Path("d_file_include"),
        Path("f_file_include"),
    }
    src = Dir(tmp_path, *items)
    src.mkdir()
    with sh.cd(tmp_path):
        contents = set(sh.ls("", include=include))
    assert contents == expected_contents


@parametrize(
    "items, kwargs, expected_contents",
    [
        param(
            [
                Dir("a_dir_include"),
                Dir("b_dir"),
                Dir("c_dir_include"),
                File("d_file_include"),
                File("e_file"),
                File("f_file_include"),
            ],
            {"include": "*_include", "only_files": True},
            {Path("d_file_include"), Path("f_file_include")},
        ),
        param(
            [
                Dir("a_dir_include"),
                Dir("b_dir"),
                Dir("c_dir_include"),
                File("d_file_include"),
                File("e_file"),
                File("f_file_include"),
            ],
            {"include": "*_include", "only_dirs": True},
            {Path("a_dir_include"), Path("c_dir_include")},
        ),
    ],
)
def test_ls__uses_only_files_and_only_dirs_in_include(
    tmp_path: Path,
    items: t.List[t.Union[Dir, File]],
    kwargs: dict,
    expected_contents: t.Set[Path],
):
    src = Dir(tmp_path, *items)
    src.mkdir()
    with sh.cd(tmp_path):
        contents = set(sh.ls("", **kwargs))
    assert contents == expected_contents


@parametrize(
    "exclude",
    [
        param("*_exclude", id="str"),
        param(["foo", "*_exclude"], id="str_list"),
        param(re.compile(fnmatch.translate("*_exclude")), id="regex"),
        param(
            [re.compile(fnmatch.translate("foo")), re.compile(fnmatch.translate("*_exclude"))],
            id="regex_list",
        ),
        param(lambda p: p.name.endswith("_exclude"), id="callable"),
        param(
            [lambda p: p.name == "foo", lambda p: p.name.endswith("_exclude")], id="callable_list"
        ),
    ],
)
def test_ls__excludes_on_multiple_types(tmp_path: Path, exclude: LsFilter):
    items: t.List[t.Union[Dir, File]] = [
        Dir("a_dir_exclude"),
        Dir("b_dir"),
        Dir("c_dir_exclude"),
        File("d_file_exclude"),
        File("e_file"),
        File("f_file_exclude"),
    ]
    expected_contents = {Path("b_dir"), Path("e_file")}
    src = Dir(tmp_path, *items)
    src.mkdir()
    with sh.cd(tmp_path):
        contents = set(sh.ls("", exclude=exclude))
    assert contents == expected_contents


def test_ls__does_not_recurse_into_excluded_dirs(tmp_path: Path):
    items: t.List[t.Union[Dir, File]] = [
        Dir("a_dir_excluded", File("a1.txt")),
        Dir("b_dir", File("b2.txt")),
    ]
    expected_contents = {
        Path("b_dir"),
        Path("b_dir/b2.txt"),
    }
    exclude = "*_excluded"
    src = Dir(tmp_path, *items)
    src.mkdir()
    with sh.cd(tmp_path):
        contents = set(sh.ls("", exclude=exclude, recursive=True))
    assert contents == expected_contents


@parametrize(
    "path, kwargs, expected",
    [
        param(".", {}, "Ls(path='.', recursive=False)"),
        param("/foo/bar/baz", {}, "Ls(path='/foo/bar/baz', recursive=False)"),
        param(".", {"recursive": True}, "Ls(path='.', recursive=True)"),
    ],
)
def test_ls__has_repr(path, kwargs, expected):
    listing = sh.ls(path, **kwargs)
    assert repr(listing) == expected


def test_ls__raises_when_both_only_files_and_only_dirs_are_true():
    with pytest.raises(ValueError):
        list(sh.ls(only_files=True, only_dirs=True))


def test_ls__raises_when_include_is_invalid_type():
    with pytest.raises(TypeError):
        list(sh.ls(include=True))


@parametrize(
    "fn, expected_kwargs",
    [
        param(sh.lsfiles, {"only_files": True}),
        param(sh.lsdirs, {"only_dirs": True}),
        param(sh.walk, {"recursive": True, "only_files": False, "only_dirs": False}),
        param(sh.walkfiles, {"recursive": True, "only_files": True, "only_dirs": False}),
        param(sh.walkdirs, {"recursive": True, "only_dirs": True, "only_files": False}),
    ],
)
def test_ls_aliases(tmp_path: Path, fn: t.Callable, expected_kwargs: dict):
    expected_kwargs["include"] = None
    expected_kwargs["exclude"] = None

    with mock.patch.object(sh.path, "ls") as mocked_ls:
        list(fn(tmp_path))

    assert mocked_ls.called
    assert mocked_ls.call_args[1] == expected_kwargs



"""The filesystem module contains utilities for interacting with the file system."""

from contextlib import contextmanager
import errno
import itertools
import os
from pathlib import Path
import random
import re
import shutil
import stat
import string
import typing as t

from .path import walk
from .types import StrPath


try:
    import fcntl
except ImportError:  # pragma: no cover
    fcntl = None  # type: ignore


try:
    from pwd import getpwnam
except ImportError:  # pragma: no cover
    getpwnam = None  # type: ignore


try:
    from grp import getgrnam
except ImportError:  # pragma: no cover
    getgrnam = None  # type: ignore


CHMOD_SYMBOLIC_PATTERN = re.compile(r"^(?P<who>[ugoa]*)(?P<op>[+\-=])(?P<perm>[ugo]|[rwxst]*)$")
CHMOD_SYMBOLIC_TABLE: t.Dict[str, int] = {
    "ur": stat.S_IRUSR,
    "uw": stat.S_IWUSR,
    "ux": stat.S_IXUSR,
    "us": stat.S_ISUID,
    "gr": stat.S_IRGRP,
    "gw": stat.S_IWGRP,
    "gx": stat.S_IXGRP,
    "gs": stat.S_ISGID,
    "or": stat.S_IROTH,
    "ow": stat.S_IWOTH,
    "ox": stat.S_IXOTH,
    "ot": stat.S_ISVTX,
    "ar": stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH,
    "aw": stat.S_IWUSR | stat.S_IWGRP | stat.S_IWOTH,
    "ax": stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH,
    "at": stat.S_ISVTX,
    "u": stat.S_IRWXU | stat.S_ISUID,
    "g": stat.S_IRWXG | stat.S_ISGID,
    "o": stat.S_IRWXO | stat.S_ISVTX,
}


def chmod(
    path: t.Union[StrPath, int],
    mode: t.Union[str, int],
    *,
    follow_symlinks: bool = True,
    recursive: bool = False,
) -> None:
    """
    Change file or directory permissions using numeric or symbolic modes.

    The mode can either be an integer, an octal number (e.g. ``0o600``), an octal string
    (e.g. ``"600"``), or a symbolic permissions string (e.g. ``"u+rw,g=r,o-rwx"``).

    The symbolic permissions string format is similar to what is accepted by the UNIX command
    ``chmod``:

    - Symbolic format: ``[ugoa...][-+=][rwxstugo...][,...]``
    - ``[ugoa...]``: Optional zero or more characters that set the user class parameter.

      - ``u``: user
      - ``g``: group
      - ``o``: other
      - ``a``: all
      - Defaults to ``a`` when none given

    - ``[-+=]``: Required operation that modifies the permissions.

      - ``-``: removes the given permissions
      - ``+``: adds the given permissions
      - ``=``: sets the given permissions to what was specified
      - If ``=`` is used without permissions, then the user class will have all of its permissions
        removed

    - ``[rwxstugo...]``: Permissions to modify for the given user classes.

      - ``r``: Read
      - ``w``: Write
      - ``x``: Execute
      - ``s``: User or Group ID bit
      - ``t``: Sticky bit
      - ``u``: User permission bits of the original path mode
      - ``g``: Group permission bits of the original path mode
      - ``o``: Other permission bits of the original path mode

    - Multiple permission clauses are separated with ``,``.

    Examples::

        # Set permissions to 600 using octal number.
        chmod(path, 0o600)

        # Set permissions to 600 using octal string.
        chmod(path, "600")

        # Set user to read-write, group to read, and remove read-write-execute from other
        chmod(path, "u=rw,g=r,o-rwx")

        # Set user, group, and other to read-write
        chmod(path, "a=rw")

        # Add execute permission for user, group, and other
        chmod(path, "+x")

        # Add user id bit, group id bit, and set sticky bit
        chmod(path, "u+s,g+s,+t")

        # Set group permission to same as user
        chmod(path, "g=u")

    Args:
        path: File, directory, or file-descriptor.
        mode: Permission mode to set.
        follow_symlinks: Whether to follow symlinks.
        recursive: Whether to recursively apply permissions to subdirectories and their files.
    """
    if not isinstance(path, int):
        path = Path(path)

    if isinstance(mode, str):
        # Attempt to convert mode from octal string to integer to support values like "640".
        try:
            mode = int(mode, 8)
        except (ValueError, TypeError):
            pass

    # Store original mode in case recursive=True since symbolic mode can depend on the target's
    # current mode so we may need a unique mode for each which will have to start with the original.
    original_mode = mode

    if isinstance(mode, str):
        # Process mode as symbolic permissions like "ug=rw,o=r".
        if isinstance(path, int):
            path_stat = os.stat(path)
        else:
            path_stat = path.stat()
        mode = _get_symbolic_mode(path_stat.st_mode, mode)

    os.chmod(path, mode, follow_symlinks=follow_symlinks)

    if recursive and isinstance(path, Path) and path.is_dir():
        for subpath in walk(path):
            # Disable recursive option so we can handle all sub-paths from here instead of using
            # recursive function calls.
            chmod(subpath, original_mode, follow_symlinks=follow_symlinks, recursive=False)


def _get_symbolic_mode(base_mode: int, symbolic_mode: str) -> int:
    """Return integer mode from a symbolic mode."""
    mode = base_mode
    items = symbolic_mode.split(",")

    for item in items:
        match = CHMOD_SYMBOLIC_PATTERN.match(item)

        if not match:
            raise ValueError(f"chmod: Unsupported symbolic mode: {symbolic_mode}")

        who = match.group("who")
        op = match.group("op")
        perm = match.group("perm")

        if not who:
            who = "a"

        mask = 0
        for who_char, perm_char in itertools.product(who, perm):
            if perm_char in "ugo":
                # Permission character is a who-class that we should inherit permissions from.
                submask = _get_inherited_symbolic_mode(mode, to_who=who_char, from_who=perm_char)
            else:
                symbol = who_char + perm_char
                if symbol not in CHMOD_SYMBOLIC_TABLE:
                    raise ValueError(f"chmod: Unsupported symbolic mode: {symbolic_mode}")
                submask = CHMOD_SYMBOLIC_TABLE[symbol]
            mask |= submask

        if op == "=":
            # Since we're setting permissions to be equal to the given mode, clear the existing
            # mode for each "who" so that its permissions will be set to just what was given.
            mode = _clear_symbolic_mode(mode, who)

        if op == "-":
            mode &= ~mask
        else:
            # Handles both "=" and "+" operators.
            mode |= mask

    return mode


def _get_inherited_symbolic_mode(base_mode: int, to_who: str, from_who: str) -> int:
    """Return integer mode by inheriting the permissions from another symbolic user class."""
    mode = 0
    for perm_char in "rwxst":
        from_symbol = from_who + perm_char
        to_symbol = to_who + perm_char
        if (
            from_symbol in CHMOD_SYMBOLIC_TABLE
            and to_symbol in CHMOD_SYMBOLIC_TABLE
            and base_mode & CHMOD_SYMBOLIC_TABLE[from_symbol]
        ):
            mode |= CHMOD_SYMBOLIC_TABLE[to_symbol]
    return mode


def _clear_symbolic_mode(mode: int, who: str) -> int:
    """Return integer mode that has been cleared for a given symbolic user class."""
    for who_char, perm_char in itertools.product(who, "rwxst"):
        symbol = who_char + perm_char
        if symbol not in CHMOD_SYMBOLIC_TABLE:
            continue
        mode &= ~CHMOD_SYMBOLIC_TABLE[who_char + perm_char]
    return mode


def chown(
    path: t.Union[StrPath, int],
    user: t.Optional[t.Union[str, int]] = None,
    group: t.Optional[t.Union[str, int]] = None,
    *,
    follow_symlinks: bool = True,
    recursive: bool = False,
) -> None:
    """
    Change ownership of file or directory to user and/or group.

    User and group can be a string name or a numeric id. Leave as ``None`` to not change the
    respective user or group ownership.

    Args:
        path: File, directory, or file-descriptor.
        user: User name or uid to set as owner. Use ``None`` or ``-1`` to not change.
        group: Group name or gid to set as owner. Use ``None`` or ``-1`` to not change.
        follow_symlinks: Whether to follow symlinks.
        recursive: Whether to recursively apply ownership to subdirectories and their files.
    """
    if user in (None, -1) and group in (None, -1):
        raise ValueError("chown: user and/or group must be set")

    if user is None:
        # -1 means don't change it
        uid = -1
    else:
        uid = _get_uid(user)  # type: ignore
        if uid is None:
            raise LookupError(f"chown: no such user: {user!r}")

    if group is None:
        # -1 means don't change it
        gid = -1
    else:
        gid = _get_gid(group)  # type: ignore
        if gid is None:
            raise LookupError(f"chown: no such group: {group!r}")

    if not isinstance(path, int):
        path = Path(path)

    os.chown(path, uid, gid, follow_symlinks=follow_symlinks)

    if recursive and isinstance(path, Path) and path.is_dir():
        for subpath in walk(path):
            # Disable recursive option so we can handle all sub-paths from here instead of using
            # recursive function calls.
            chown(subpath, uid, gid, follow_symlinks=follow_symlinks, recursive=False)


def _get_uid(name: t.Optional[t.Union[str, int]]) -> t.Optional[int]:
    """Return an uid given a user name."""
    uid: t.Optional[int] = None
    if isinstance(name, int):
        uid = name
    elif isinstance(name, str):
        try:
            uid = getpwnam(name).pw_uid
        except (KeyError, TypeError, ValueError):
            pass
    return uid


def _get_gid(name: t.Optional[t.Union[str, int]]) -> t.Optional[int]:
    """Return a gid given a group name."""
    gid: t.Optional[int] = None
    if isinstance(name, int):
        gid = name
    elif isinstance(name, str):
        try:
            gid = getgrnam(name).gr_gid
        except (KeyError, TypeError, ValueError):
            pass
    return gid


def cp(src: StrPath, dst: StrPath, *, follow_symlinks: bool = True) -> None:
    """
    Copy file or directory to destination.

    Files are copied atomically by first copying to a temporary file in the same target directory
    and then renaming the temporary file to its actual filename.

    Args:
        src: Source file or directory to copy from.
        dst: Destination file or directory to copy to.
        follow_symlinks: When true (the default), symlinks in the source will be dereferenced into
            the destination. When false, symlinks in the source will be preserved as symlinks in the
            destination.
    """
    src = Path(src)
    dst = Path(dst)
    mkdir(dst.parent)

    if src.is_dir():
        if dst.exists() and not dst.is_dir():
            raise FileExistsError(
                errno.EEXIST, f"Cannot copy {src!r} to {dst!r} since destination is a file"
            )

        if dst.is_dir():
            src_dirname = str(src)
            dst_dirname = str(dst)
            for src_dir, _dirs, files in os.walk(str(src)):
                dst_dir = src_dir.replace(src_dirname, dst_dirname, 1)
                for file in files:
                    src_file = os.path.join(src_dir, file)
                    dst_file = os.path.join(dst_dir, file)
                    cp(src_file, dst_file, follow_symlinks=follow_symlinks)
        else:

            def copy_function(_src, _dst):
                return cp(_src, _dst, follow_symlinks=follow_symlinks)

            shutil.copytree(src, dst, symlinks=not follow_symlinks, copy_function=copy_function)
    else:
        if dst.is_dir():
            dst = dst / src.name
        tmp_dst = _candidate_temp_pathname(path=dst, prefix="_")
        shutil.copy2(src, tmp_dst, follow_symlinks=follow_symlinks)
        try:
            os.rename(tmp_dst, dst)
        except OSError:  # pragma: no cover
            rm(tmp_dst)
            raise


def dirsync(path: StrPath) -> None:
    """
    Force sync on directory.

    Args:
        path: Directory to sync.
    """
    fd = os.open(path, os.O_RDONLY)
    try:
        fsync(fd)
    finally:
        os.close(fd)


@contextmanager
def environ(
    env: t.Optional[t.Dict[str, str]] = None, *, replace: bool = False
) -> t.Iterator[t.Dict[str, str]]:
    """
    Context manager that updates environment variables with `env` on enter and restores the original
    environment on exit.

    Args:
        env: Environment variables to set.
        replace: Whether to clear existing environment variables before setting new ones. This fully
            replaces the existing environment variables so that only `env` are set.

    Yields:
        The current environment variables.
    """
    orig_env = os.environ.copy()

    if replace:
        os.environ.clear()

    if env:
        os.environ.update(env)

    try:
        yield os.environ.copy()
    finally:
        os.environ.clear()
        os.environ.update(orig_env)


def fsync(fd: t.Union[t.IO, int]) -> None:
    """
    Force write of file to disk.

    The file descriptor will have ``os.fsync()`` (or ``fcntl.fcntl()`` with ``fcntl.F_FULLFSYNC``
    if available) called on it. If a file object is passed it, then it will first be flushed before
    synced.

    Args:
        fd: Either file descriptor integer or file object.
    """
    if (
        not isinstance(fd, int)
        and not (hasattr(fd, "fileno") and hasattr(fd, "flush"))
        or isinstance(fd, bool)
    ):
        raise ValueError(
            f"File descriptor must be a fileno integer or file-like object, not {type(fd)}"
        )

    if isinstance(fd, int):
        fileno = fd
    else:
        fileno = fd.fileno()
        fd.flush()

    if hasattr(fcntl, "F_FULLFSYNC"):  # pragma: no cover
        # Necessary for MacOS to do proper fsync: https://bugs.python.org/issue11877
        fcntl.fcntl(fileno, fcntl.F_FULLFSYNC)  # pylint: disable=no-member
    else:  # pragma: no cover
        os.fsync(fileno)


def getdirsize(path: StrPath, pattern: str = "**/*") -> int:
    """
    Return total size of directory's contents.

    Args:
        path: Directory to calculate total size of.
        pattern: Only count files if they match this glob-pattern.

    Returns:
        Total size of directory in bytes.
    """
    total_size = 0

    for item in Path(path).glob(pattern):
        if item.is_file():
            try:
                total_size += item.stat().st_size
            except OSError:  # pragma: no cover
                # File doesn't exist or is inaccessible.
                pass

    return total_size


def mkdir(*paths: StrPath, mode: int = 0o777, exist_ok: bool = True) -> None:
    """
    Recursively create directories in `paths` along with any parent directories that don't already
    exists.

    This is like the Unix command ``mkdir -p <path1> <path2> ...``.

    Args:
        *paths: Directories to create.
        mode: Access mode for directories.
        exist_ok: Whether it's ok or not if the path already exists. When ``True``, a
            ``FileExistsError`` will be raised.
    """
    for path in paths:
        os.makedirs(path, mode=mode, exist_ok=exist_ok)


def mv(src: StrPath, dst: StrPath) -> None:
    """
    Move source file or directory to destination.

    The move semantics are as follows:

    - If src and dst are files, then src will be renamed to dst and overwrite dst if it exists.
    - If src is a file and dst is a directory, then src will be moved under dst.
    - If src is a directory and dst does not exist, then src will be renamed to dst and any parent
      directories that don't exist in the dst path will be created.
    - If src is a directory and dst is a directory and the src's basename does not exist under dst
      or if it is an empty directory, then src will be moved under dst.
    - If src is directory and dst is a directory and the src's basename is a non-empty directory
      under dst, then an ``OSError`` will be raised.
    - If src and dst reference two difference file-systems, then src will be copied to dst using
      :func:`.cp` and then deleted at src.

    Args:
        src: Source file or directory to move.
        dst: Destination file or directory to move source to.
    """
    src = Path(src)
    dst = Path(dst)
    mkdir(dst.parent)

    if dst.is_dir():
        dst = dst / src.name

    try:
        os.rename(src, dst)
    except OSError as exc:
        if exc.errno == errno.EXDEV:
            # errno.EXDEV means we tried to move from one file-system to another which is not
            # allowed. In that case, we'll fallback to a copy-and-delete approach instead.
            tmp_dst = _candidate_temp_pathname(path=dst, prefix="_")
            try:
                cp(src, tmp_dst)
                os.rename(tmp_dst, dst)
                rm(src)
            finally:
                rm(tmp_dst)
        else:
            raise


def rm(*paths: StrPath) -> None:
    """
    Delete files and directories.

    Note:
        Deleting non-existent files or directories does not raise an error.

    Warning:
        This function is like ``$ rm -rf`` so be careful. To limit the scope of the removal to just
        files or just directories, use :func:`.rmfile` or :func:`.rmdir` respectively.

    Args:
        *paths: Files and/or directories to delete.
    """
    for path in paths:
        try:
            try:
                shutil.rmtree(path)
            except NotADirectoryError:
                os.remove(path)
        except FileNotFoundError:
            pass


def rmdir(*dirs: StrPath) -> None:
    """
    Delete directories.

    Note:
        Deleting non-existent directories does not raise an error.

    Warning:
        This function is like calling ``$ rm -rf`` on a directory. To limit the scope of the removal
        to just files, use :func:`.rmfile`.

    Args:
        *dirs: Directories to delete.

    Raises:
        NotADirectoryError: When given path is not a directory.
    """
    for path in dirs:
        try:
            shutil.rmtree(path)
        except FileNotFoundError:
            pass


def rmfile(*files: StrPath) -> None:
    """
    Delete files.

    Note:
        Deleting non-existent files does not raise an error.

    Args:
        *files: Files to delete.

    Raises:
        IsADirectoryError: When given path is a directory.
    """
    for path in files:
        try:
            os.remove(path)
        except FileNotFoundError:
            pass


def touch(*paths: StrPath) -> None:
    """
    Touch files.

    Args:
        *paths: File paths to create.
    """
    for path in paths:
        path = Path(path)
        mkdir(path.parent)
        path.touch()


@contextmanager
def umask(mask: int = 0) -> t.Iterator[None]:
    """
    Context manager that sets the umask to `mask` and restores it on exit.

    Args:
        mask: Numeric umask to set.

    Yields:
        None
    """
    orig_mask = os.umask(mask)
    try:
        yield
    finally:
        os.umask(orig_mask)


def _candidate_temp_pathname(
    path: StrPath = "", prefix: StrPath = "", suffix: StrPath = "", hidden: bool = True
) -> str:
    """Return random temporary path name that doesn't yet exist."""
    tries = 100
    for _ in range(tries):
        filename = Path(_random_name(path=path, prefix=prefix, suffix=suffix))
        if hidden:
            filename = filename.parent / f".{filename.name}"
        if not filename.exists():
            return str(filename)
    raise FileNotFoundError(
        errno.ENOENT, f"No usable temporary filename found in {Path(prefix).absolute()}"
    )  # pragma: no cover


def _random_name(
    path: StrPath = "", prefix: StrPath = "", suffix: StrPath = "", length: int = 8
) -> str:
    """Return generated random path name."""
    _pid, _random = getattr(_random_name, "_state", (None, None))
    if _pid != os.getpid() or not _random:
        # Ensure separate processes don't share same random generator.
        _random = random.Random()
        _random_name._state = (os.getpid(), _random)  # type: ignore

    inner = "".join(_random.choice(string.ascii_letters) for _ in range(length))
    return f"{path}{prefix}{inner}{suffix}"



"""The path module contains utilities for working with OS paths."""

from contextlib import contextmanager
import fnmatch
import os
from pathlib import Path
import typing as t
from typing import Iterable

from .types import LsFilter, LsFilterable, LsFilterFn, StrPath


class Ls:
    """
    Directory listing iterable that iterates over its contents and returns them as ``Path`` objects.

    Args:
        path: Directory to list.
        recursive: Whether to recurse into subdirectories. Defaults to ``False``.
        only_files: Limit results to files only. Mutually exclusive with ``only_dirs``.
        only_dirs: Limit results to directories only. Mutually exclusive with ``only_files``.
        include: Include paths by filtering on a glob-pattern string, compiled regex, callable, or
            iterable containing any of those types. Path is included if any of the filters return
            ``True`` and path matches ``only_files`` or ``only_dirs`` (if set). If path is a
            directory and is not included, its contents are still eligible for inclusion if they
            match one of the include filters.
        exclude: Exclude paths by filtering on a glob-pattern string, compiled regex, callable, or
            iterable containing any of those types. Path is not yielded if any of the filters return
            ``True``. If the path is a directory and is excluded, then all of its contents will be
            excluded.
    """

    def __init__(
        self,
        path: StrPath = ".",
        *,
        recursive: bool = False,
        only_files: bool = False,
        only_dirs: bool = False,
        include: t.Optional[LsFilter] = None,
        exclude: t.Optional[LsFilter] = None,
    ):
        if only_files and only_dirs:
            raise ValueError("only_files and only_dirs cannot both be true")

        include_filters: t.List[t.Callable[[Path], bool]] = []
        exclude_filters: t.List[t.Callable[[Path], bool]] = []

        if include:
            if isinstance(include, Iterable) and not isinstance(include, (str, bytes)):
                includes = include
            else:
                includes = [include]
            # When creating the include filters, need to also take into account the only_* filter
            # settings so that an include filter will only match if both are true.
            include_filters.extend(
                _make_ls_filter(only_files=only_files, only_dirs=only_dirs, filterable=incl)
                for incl in includes
            )
        elif only_files or only_dirs:
            # If no include filters are given, but one of the only_* filters is, then we'll add it.
            # Otherwise, when include is given, the only_* filters are taken into account for each
            # include filter.
            include_filters.append(_make_ls_filter(only_files=only_files, only_dirs=only_dirs))

        if exclude:
            if isinstance(exclude, Iterable) and not isinstance(exclude, (str, bytes)):
                excludes = exclude
            else:
                excludes = [exclude]
            exclude_filters.extend(_make_ls_filter(filterable=excl) for excl in excludes)

        self.path = path
        self.recursive = recursive
        self.include_filters = include_filters
        self.exclude_filters = exclude_filters

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(path={self.path!r}, recursive={self.recursive})"

    def __str__(self) -> str:
        """Return file system representation of path."""
        return str(self.path)

    # Make class PathLike for os.fspath compatibility.
    __fspath__ = __str__

    def __iter__(self) -> t.Iterator[Path]:
        """Iterate over :attr:`path` and yield its contents."""
        yield from _ls(
            self.path,
            recursive=self.recursive,
            include_filters=self.include_filters,
            exclude_filters=self.exclude_filters,
        )


def _ls(
    path: StrPath = ".",
    *,
    recursive: bool = False,
    include_filters: t.Optional[t.List[LsFilterFn]] = None,
    exclude_filters: t.Optional[t.List[LsFilterFn]] = None,
) -> t.Generator[Path, None, None]:
    scanner = os.scandir(Path(path))
    recurse_into: t.List[str] = []

    with scanner:
        while True:
            try:
                try:
                    entry = next(scanner)
                except StopIteration:
                    break
            except OSError:  # pragma: no cover
                return

            entry_path = Path(entry.path)
            excluded = exclude_filters and any(
                is_excluded(entry_path) for is_excluded in exclude_filters
            )

            if not excluded and (
                not include_filters
                or any(is_included(entry_path) for is_included in include_filters)
            ):
                yield entry_path

            if recursive and not excluded and entry_path.is_dir() and not entry_path.is_symlink():
                recurse_into.append(entry.path)

    for subdir in recurse_into:
        yield from _ls(
            subdir,
            recursive=recursive,
            include_filters=include_filters,
            exclude_filters=exclude_filters,
        )


def _make_ls_filter(
    only_files: bool = False, only_dirs: bool = False, filterable: t.Optional[LsFilterable] = None
) -> LsFilterFn:
    filter_fn: t.Optional[LsFilterFn] = None
    if filterable:
        filter_fn = _make_ls_filterable_fn(filterable)

    def _ls_filter(path: Path) -> bool:
        if only_files and path.is_dir():
            return False
        elif only_dirs and path.is_file():
            return False
        elif filter_fn:
            return filter_fn(path)
        else:
            return True

    return _ls_filter


def _make_ls_filterable_fn(filterable: LsFilterable) -> LsFilterFn:
    _ls_filterable_fn: LsFilterFn

    if isinstance(filterable, str):

        def _ls_filterable_fn(path: Path) -> bool:
            return fnmatch.fnmatch(path, filterable)  # type: ignore

    elif isinstance(filterable, t.Pattern):

        def _ls_filterable_fn(path: Path) -> bool:
            return bool(filterable.match(str(path)))  # type: ignore

    elif callable(filterable):

        def _ls_filterable_fn(path: Path) -> bool:
            return filterable(path)  # type: ignore

    else:
        raise TypeError(
            f"ls filter must be one of str, re.compile() or callable, not {type(filterable)!r}"
        )

    return _ls_filterable_fn


@contextmanager
def cd(path: StrPath) -> t.Iterator[None]:
    """
    Context manager that changes the working directory on enter and restores it on exit.

    Args:
        path: Directory to change to.
    """
    orig_cwd = os.getcwd()

    if path:
        os.chdir(path)

    try:
        yield
    finally:
        if path:
            os.chdir(orig_cwd)


def cwd() -> Path:
    """Return current working directory as ``Path`` object."""
    return Path.cwd()


def homedir():
    """Return current user's home directory as ``Path`` object."""
    return Path.home()


def ls(
    path: StrPath = ".",
    *,
    recursive: bool = False,
    only_files: bool = False,
    only_dirs: bool = False,
    include: t.Optional[LsFilter] = None,
    exclude: t.Optional[LsFilter] = None,
) -> Ls:
    """
    Return iterable that lists directory contents as ``Path`` objects.

    Args:
        path: Directory to list.
        recursive: Whether to recurse into subdirectories. Defaults to ``False``.
        only_files: Limit results to files only. Mutually exclusive with ``only_dirs``.
        only_dirs: Limit results to directories only. Mutually exclusive with ``only_files``.
        include: Include paths by filtering on a glob-pattern string, compiled regex, callable, or
            iterable containing any of those types. Path is included if any of the filters return
            ``True`` and path matches ``only_files`` or ``only_dirs`` (if set). If path is a
            directory and is not included, its contents are still eligible for inclusion if they
            match one of the include filters.
        exclude: Exclude paths by filtering on a glob-pattern string, compiled regex, callable, or
            iterable containing any of those types. Path is not yielded if any of the filters return
            ``True``. If the path is a directory and is excluded, then all of its contents will be
            excluded.
    """
    return Ls(
        path,
        recursive=recursive,
        only_files=only_files,
        only_dirs=only_dirs,
        include=include,
        exclude=exclude,
    )


def lsfiles(
    path: StrPath = ".",
    *,
    include: t.Optional[LsFilter] = None,
    exclude: t.Optional[LsFilter] = None,
) -> Ls:
    """
    Return iterable that only lists files in directory as ``Path`` objects.

    See Also:
        This function is not recursive and will only yield the top-level contents of a directory.
        Use :func:`.walkfiles` to recursively yield all files from a directory.

    Args:
        path: Directory to list.
        include: Include paths by filtering on a glob-pattern string, compiled regex, callable, or
            iterable containing any of those types. Path is included if any of the filters return
            ``True``. If path is a directory and is not included, its contents are still eligible
            for inclusion if they match one of the include filters.
        exclude: Exclude paths by filtering on a glob-pattern string, compiled regex, callable, or
            iterable containing any of those types. Path is not yielded if any of the filters return
            ``True``. If the path is a directory and is excluded, then all of its contents will be
            excluded.
    """
    return ls(path, only_files=True, include=include, exclude=exclude)


def lsdirs(
    path: StrPath = ".",
    *,
    include: t.Optional[LsFilter] = None,
    exclude: t.Optional[LsFilter] = None,
) -> Ls:
    """
    Return iterable that only lists directories in directory as ``Path`` objects.

    See Also:
        This function is not recursive and will only yield the top-level contents of a directory.
        Use :func:`.walkdirs` to recursively yield all directories from a directory.

    Args:
        path: Directory to list.
        include: Include paths by filtering on a glob-pattern string, compiled regex, callable, or
            iterable containing any of those types. Path is included if any of the filters return
            ``True``. If path is a directory and is not included, its contents are still eligible
            for inclusion if they match one of the include filters.
        exclude: Exclude paths by filtering on a glob-pattern string, compiled regex, callable, or
            iterable containing any of those types. Path is not yielded if any of the filters return
            ``True``. If the path is a directory and is excluded, then all of its contents will be
            excluded.
    """
    return ls(path, only_dirs=True, include=include, exclude=exclude)


def reljoin(*paths: StrPath) -> str:
    """
    Like ``os.path.join`` except that all paths are treated as relative to the previous one so that
    an absolute path in the middle will extend the existing path instead of becoming the new root
    path.

    Args:
        *paths: Paths to join together.
    """
    path = os.sep.join(str(Path(path)) for path in paths)
    return os.path.normpath(path)


def walk(
    path: StrPath = ".",
    *,
    only_files: bool = False,
    only_dirs: bool = False,
    include: t.Optional[LsFilter] = None,
    exclude: t.Optional[LsFilter] = None,
) -> Ls:
    """
    Return iterable that recursively lists all directory contents as ``Path`` objects.

    See Also:
        This function is recursive and will list all contents of a directory. Use :func:`.ls` to
        list only the top-level contents of a directory.

    Args:
        path: Directory to walk.
        only_files: Limit results to files only. Mutually exclusive with ``only_dirs``.
        only_dirs: Limit results to directories only. Mutually exclusive with ``only_files``.
        include: Include paths by filtering on a glob-pattern string, compiled regex, callable, or
            iterable containing any of those types. Path is included if any of the filters return
            ``True`` and path matches ``only_files`` or ``only_dirs`` (if set). If path is a
            directory and is not included, its contents are still eligible for inclusion if they
            match one of the include filters.
        exclude: Exclude paths by filtering on a glob-pattern string, compiled regex, callable, or
            iterable containing any of those types. Path is not yielded if any of the filters return
            ``True``. If the path is a directory and is excluded, then all of its contents will be
            excluded.
    """
    return ls(
        path,
        recursive=True,
        only_files=only_files,
        only_dirs=only_dirs,
        include=include,
        exclude=exclude,
    )


def walkfiles(
    path: StrPath = ".",
    *,
    include: t.Optional[LsFilter] = None,
    exclude: t.Optional[LsFilter] = None,
) -> Ls:
    """
    Return iterable that recursively lists only files in directory as ``Path`` objects.

    See Also:
        This function is recursive and will list all files in a directory. Use :func:`.lsfiles` to
        list only the top-level files in a directory.

    Args:
        path: Directory to walk.
        include: Include paths by filtering on a glob-pattern string, compiled regex, callable, or
            iterable containing any of those types. Path is included if any of the filters return
            ``True``. If path is a directory and is not included, its contents are still eligible
            for inclusion if they match one of the include filters.
        exclude: Exclude paths by filtering on a glob-pattern string, compiled regex, callable, or
            iterable containing any of those types. Path is not yielded if any of the filters return
            ``True``. If the path is a directory and is excluded, then all of its contents will be
            excluded.
    """
    return walk(path, only_files=True, include=include, exclude=exclude)


def walkdirs(
    path: StrPath = ".",
    *,
    include: t.Optional[LsFilter] = None,
    exclude: t.Optional[LsFilter] = None,
) -> Ls:
    """
    Return iterable that recursively lists only directories in directory as ``Path`` objects.

    See Also:
        This function is recursive and will list all directories in a directory. Use
        :func:`.lsfiles` to list only the top-level directories in a directory.

    Args:
        path: Directory to walk.
        include: Include paths by filtering on a glob-pattern string, compiled regex, callable, or
            iterable containing any of those types. Path is included if any of the filters return
            ``True``. If path is a directory and is not included, its contents are still eligible
            for inclusion if they match one of the include filters.
        exclude: Exclude paths by filtering on a glob-pattern string, compiled regex, callable, or
            iterable containing any of those types. Path is not yielded if any of the filters return
            ``True``. If the path is a directory and is excluded, then all of its contents will be
            excluded.
    """
    return walk(path, only_dirs=True, include=include, exclude=exclude)



import os
from pathlib import Path

import pytest
from pytest import param

import shelmet as sh


parametrize = pytest.mark.parametrize


@parametrize(
    "path",
    [
        param(""),
        param("a"),
        param("a/b"),
        param("a/b/c"),
    ],
)
def test_cd__changes_cwd(tmp_path: Path, path: str):
    orig_cwd = os.getcwd()
    cd_path = tmp_path / path
    cd_path.mkdir(parents=True, exist_ok=True)

    with sh.cd(cd_path):
        assert os.getcwd() == str(cd_path)
    assert os.getcwd() == orig_cwd



import errno
import os
from pathlib import Path
import typing as t
from unittest import mock

import pytest
from pytest import param

import shelmet as sh

from .utils import Dir, File


parametrize = pytest.mark.parametrize


@parametrize(
    "src_file, dst_file, expected",
    [
        param(
            File("src.txt", text="src"),
            File("dst.txt"),
            File("dst.txt", text="src"),
            id="to_new_file",
        ),
        param(
            File("src.txt", text="src"),
            File("dst.txt", text="dst"),
            File("dst.txt", text="src"),
            id="to_overwite_existing_file",
        ),
    ],
)
def test_mv__moves_file_to_file(tmp_path: Path, src_file: File, dst_file: File, expected: File):
    base_dir = Dir(tmp_path, src_file)
    base_dir.mkdir()

    expected_file_path = base_dir.path / expected.path
    expected_file_text = expected.text

    dst_target_path = base_dir.path / dst_file.path
    if dst_file.text:
        dst_target_path.write_text(dst_file.text)

    sh.mv(src_file.path, dst_target_path)

    assert not src_file.path.exists()
    assert expected_file_path.exists()
    assert expected_file_path.read_text() == expected_file_text


@parametrize(
    "src_file, dst_dir, expected",
    [
        param(
            File("src.txt", text="src"),
            Dir("dst"),
            File("dst/src.txt", text="src"),
            id="to_new_file_under_destination",
        ),
    ],
)
def test_mv__moves_file_to_dir(tmp_path: Path, src_file: File, dst_dir: Dir, expected: File):
    base_dir = Dir(tmp_path, src_file)
    base_dir.mkdir()

    expected_file_path = base_dir.path / expected.path
    expected_file_text = expected.text

    dst_target_path = base_dir.path / dst_dir.path
    dst_target_path.mkdir()

    sh.mv(src_file.path, dst_target_path)

    assert not src_file.path.exists()
    assert expected_file_path.exists()
    assert expected_file_path.read_text() == expected_file_text


@parametrize(
    "src_files, dst, expected",
    [
        param([File("1.txt", text="1")], "dst", "dst", id="to_new_dir"),
        param(
            [File("1.txt", text="1")],
            Dir("dst", File("2.txt")),
            "dst/src",
            id="to_new_dir_under_destination",
        ),
        param(
            [File("1.txt", text="1")],
            Dir("dst", Dir("src")),
            "dst/src",
            id="to_new_dir_overwriting_existing_dir_under_destination",
        ),
    ],
)
def test_mv__moves_dir(
    tmp_path: Path, src_files: t.List[File], dst: t.Union[Dir, str], expected: str
):
    src_dir = Dir(tmp_path / "src", *src_files)
    src_dir.mkdir()

    if isinstance(dst, Dir):
        dst_dir = Dir(tmp_path / dst.path, *dst.files)
        dst_dir.mkdir()
    else:
        dst_dir = Dir(tmp_path / dst)

    sh.mv(src_dir.path, dst_dir.path)

    expected_dst_dir = Dir(tmp_path / expected)
    assert not src_dir.path.exists()
    assert expected_dst_dir.path.exists()

    for src_file in src_files:
        dst_file_path = expected_dst_dir.path / src_file.path.name
        assert dst_file_path.read_text() == src_file.text


def test_mv__allows_same_file_as_destination(tmp_path: Path):
    src_file = File(tmp_path / "src.txt", text="src")
    src_file.write()
    sh.mv(src_file.path, src_file.path)
    assert src_file.path.exists()
    assert src_file.path.read_text() == src_file.text


def test_mv__works_across_file_systems(tmp_path: Path):
    src_file = File(tmp_path / "src.txt", text="src")
    src_file.write()

    dst_file = File(tmp_path / "dst.txt")
    _os_rename = os.rename

    def mock_os_rename(src, dst):
        if str(src) == str(src_file.path) and str(dst) == str(dst_file.path):
            raise OSError(errno.EXDEV, "mock error from move across file systems")
        return _os_rename(src, dst)

    with mock.patch("os.rename", side_effect=mock_os_rename):
        sh.mv(src_file.path, dst_file.path)

    assert dst_file.path.exists()
    assert dst_file.path.read_text() == src_file.text
    assert not src_file.path.exists()


def test_mv__raises_when_source_dir_exists_in_destination_and_is_not_empty(tmp_path: Path):
    src_dir = Dir(tmp_path / "src", File("src.txt", text="src"))
    src_dir.mkdir()
    dst_dir = Dir(tmp_path / "dst", File("src/other.txt", text="other"))
    dst_dir.mkdir()

    with pytest.raises(OSError):
        sh.mv(src_dir.path, dst_dir.path)



import os
from pathlib import Path
import typing as t
from uuid import uuid4

import pytest
from pytest import param

import shelmet as sh


parametrize = pytest.mark.parametrize


def lines_as_text(items: t.Iterable) -> t.List[str]:
    lines = os.linesep.join(str(item) for item in items)
    return lines.splitlines(keepends=True)


def lines_as_bytes(items: t.Iterable) -> t.List[bytes]:
    lines = os.linesep.join(str(item) for item in items).encode()
    return lines.splitlines(keepends=True)


def repeat_items(items: t.List[str], times: int) -> t.List[str]:
    return [item * times for item in items]


@pytest.fixture()
def write_text(tmp_path: Path) -> t.Callable[[str], Path]:
    def _write_text(content: str) -> Path:
        filename = tmp_path / uuid4().hex
        filename.write_text(content)
        return filename

    return _write_text


@pytest.fixture()
def write_bytes(tmp_path: Path) -> t.Callable[[bytes], Path]:
    def _write_bytes(content: bytes) -> Path:
        filename = tmp_path / uuid4().hex
        filename.write_bytes(content)
        return filename

    return _write_bytes


@pytest.fixture(
    params=[
        param("r"),
        param("rt"),
        param("tr"),
        param("rb"),
        param("br"),
    ]
)
def valid_read_only_mode(request):
    return request.param


@pytest.fixture(
    params=[
        param("r+"),
        param("rb+"),
        param("w"),
        param("w+"),
        param("wb"),
        param("wb+"),
        param("x"),
        param("a"),
        param("a+"),
    ]
)
def invalid_read_only_mode(request):
    return request.param


def test_read__returns_text_file_contents(write_text: t.Callable[[str], Path]):
    content = "some text"
    test_file = write_text(content)
    assert sh.read(test_file) == content


def test_read__returns_binary_file_contents(write_bytes: t.Callable[[bytes], Path]):
    content = b"some data"
    test_file = write_bytes(content)
    assert sh.read(test_file, "rb") == content


def test_read__accepts_valid_mode(tmp_path: Path, valid_read_only_mode):
    test_file = tmp_path / "test_file"
    test_file.touch()
    sh.read(test_file, valid_read_only_mode)


def test_read__raises_when_mode_invalid(
    write_text: t.Callable[[str], Path], invalid_read_only_mode: str
):
    test_file = write_text("test")
    with pytest.raises(ValueError):
        sh.read(test_file, invalid_read_only_mode)


def test_readtext__returns_text_file_contents(write_text: t.Callable[[str], Path]):
    content = "some text"
    test_file = write_text(content)
    assert sh.readtext(test_file) == content


def test_readbytes__returns_binary_file_contents(write_bytes: t.Callable[[bytes], Path]):
    content = b"some data"
    test_file = write_bytes(content)
    assert sh.readbytes(test_file) == content


@parametrize(
    "chunks, size, sep",
    [
        param(repeat_items(["1", "2", "3", "4"], 1), 1, None),
        param(repeat_items(["1", "2", "3", "4"], 2), 2, None),
        param(repeat_items(["1", "2", "3", "4"], 50), 50, None),
        param(repeat_items(["1", "2", "3", "4"], 1), 1, "|"),
        param(repeat_items(["1", "2", "3", "4"], 1), 4, "|"),
        param(repeat_items(["1", "2", "3", "4"], 2), 1, "|"),
        param(repeat_items(["1", "2", "3", "4"], 2), 2, "|"),
        param(repeat_items(["1", "2", "3", "4"], 2), 4, "|"),
        param(repeat_items(["1", "2", "3", "4"], 50), 1, "|"),
        param(repeat_items(["1", "2", "3", "4"], 50), 5, "|"),
        param(repeat_items(["1", "2", "3", "4"], 50), 25, "|"),
        param(repeat_items(["1", "2", "3", "4"], 50), 1, ";|&"),
        param(repeat_items(["1", "2", "3", "4"], 50), 25, ";|&"),
        param(repeat_items(["1", "2", "3", "4"], 50), 50, ";|&"),
        param(repeat_items(["1", "2", "3", "4"], 50), 100, ";|&"),
    ],
)
def test_readchunks__yields_text_chunks_by_size(
    write_text: t.Callable[[str], Path], chunks: list, size: int, sep: str
):
    content = (sep or "").join(chunks)
    test_file = write_text(content)

    for i, chunk in enumerate(sh.readchunks(test_file, size=size, sep=sep)):
        assert chunk == chunks[i]


@parametrize(
    "chunks, size, sep",
    [
        param(repeat_items(["1", "2", "3", "4"], 1), 1, None),
        param(repeat_items(["1", "2", "3", "4"], 2), 2, None),
        param(repeat_items(["1", "2", "3", "4"], 50), 50, None),
        param(repeat_items(["1", "2", "3", "4"], 1), 1, "|"),
        param(repeat_items(["1", "2", "3", "4"], 1), 4, "|"),
        param(repeat_items(["1", "2", "3", "4"], 2), 1, "|"),
        param(repeat_items(["1", "2", "3", "4"], 2), 2, "|"),
        param(repeat_items(["1", "2", "3", "4"], 2), 4, "|"),
        param(repeat_items(["1", "2", "3", "4"], 50), 1, "|"),
        param(repeat_items(["1", "2", "3", "4"], 50), 5, "|"),
        param(repeat_items(["1", "2", "3", "4"], 50), 25, "|"),
        param(repeat_items(["1", "2", "3", "4"], 50), 1, ";|&"),
        param(repeat_items(["1", "2", "3", "4"], 50), 25, ";|&"),
        param(repeat_items(["1", "2", "3", "4"], 50), 50, ";|&"),
        param(repeat_items(["1", "2", "3", "4"], 50), 100, ";|&"),
    ],
)
def test_readchunks__yields_binary_chunks_by_size(
    write_bytes: t.Callable[[bytes], Path], chunks: list, size: int, sep: str
):
    content = (sep or "").join(chunks)
    test_file = write_bytes(content.encode())
    bin_sep: t.Optional[bytes] = sep.encode() if sep else None

    for i, chunk in enumerate(sh.readchunks(test_file, "rb", size=size, sep=bin_sep)):
        assert chunk.decode() == chunks[i]


def test_readchunks__raises_when_mode_invalid(
    write_text: t.Callable[[str], Path], invalid_read_only_mode: str
):
    test_file = write_text("test")
    with pytest.raises(ValueError):
        sh.readchunks(test_file, invalid_read_only_mode)


def test_readlines__yields_each_line_from_text_file(write_text: t.Callable[[str], Path]):
    lines = lines_as_text(range(10))
    test_file = write_text("".join(lines))

    for i, line in enumerate(sh.readlines(test_file)):
        assert line == lines[i]


def test_readlines__yields_each_line_from_binary_file(write_bytes: t.Callable[[bytes], Path]):
    lines = lines_as_bytes(range(10))
    test_file = write_bytes(b"".join(lines))

    for i, line in enumerate(sh.readlines(test_file, "rb")):
        assert line == lines[i]


def test_readlines__raises_when_mode_invalid(
    write_text: t.Callable[[str], Path], invalid_read_only_mode: str
):
    test_file = write_text("test")
    with pytest.raises(ValueError):
        sh.readlines(test_file, invalid_read_only_mode)



from contextlib import ExitStack
from pathlib import Path
import typing as t
from unittest import mock

import pytest
from pytest import param

import shelmet as sh

from .utils import (
    ARCHIVE_EXTENSIONS,
    Dir,
    File,
    create_archive,
    create_archive_source,
    create_unsafe_archive,
    is_same_dir,
)


parametrize = pytest.mark.parametrize


def _test_unarchive(tmp_path: Path, archive_file: Path, source: t.Union[File, Dir], ext: str = ""):
    src_dir = create_archive_source(tmp_path, source)
    create_archive(archive_file, src_dir.path, ext=ext)

    dst_path = tmp_path / "dst"
    sh.unarchive(archive_file, dst_path, ext=ext)

    assert dst_path.is_dir()
    assert is_same_dir(src_dir.path, dst_path / "src")


@pytest.fixture(params=ARCHIVE_EXTENSIONS)
def arc_ext(request) -> str:
    """Fixture that yields all archive extensions."""
    return request.param


@pytest.fixture(params=[".tar", ".zip"])
def rep_ext(request) -> str:
    """Fixture that yields a representative sample of archive extensions."""
    return request.param


@parametrize(
    "source", [param(Dir("a", Dir("b"), File("1.txt", text="1"), File("2.txt", text="2")))]
)
def test_unarchive__unarchives(tmp_path: Path, arc_ext: str, source: t.Union[File, Dir]):
    archive_file = tmp_path / f"archive{arc_ext}"
    _test_unarchive(tmp_path, archive_file, source)


def test_unarchive__unarchives_with_explicit_extension_format(tmp_path: Path, arc_ext: str):
    source = Dir("a", Dir("b"), File("1.txt", text="1"), File("2.txt", text="2"))
    archive_file = tmp_path / "archive"
    _test_unarchive(tmp_path, archive_file, source, ext=arc_ext)


def test_unarchive__raises_when_file_extension_not_supported():
    with pytest.raises(NotImplementedError) as exc_info:
        sh.unarchive("test.txt")
    assert "format not supported" in str(exc_info.value)


def test_unarchive__raises_when_extraction_fails(tmp_path: Path, rep_ext: str):
    archive_file = tmp_path / f"archive{rep_ext}"
    src_dir = create_archive_source(tmp_path, File("1.txt", text="1"))
    create_archive(archive_file, src_dir.path)

    with ExitStack() as mock_stack:
        mock_stack.enter_context(mock.patch("tarfile.TarFile.extractall", side_effect=Exception))
        mock_stack.enter_context(mock.patch("zipfile.ZipFile.extractall", side_effect=Exception))

        with pytest.raises(sh.ArchiveError):
            sh.unarchive(archive_file, tmp_path / "dst")


def test_unarchive__unarchives_trusted_archive_outside_target(tmp_path: Path):
    src_dir = create_archive_source(tmp_path, File("1.txt", text="1"))

    unsafe_archive_file = tmp_path / "unsafe.tar"
    unsafe_dest = tmp_path / "unsafe"
    create_unsafe_archive(unsafe_archive_file, src_dir.path, unsafe_dest)

    dst_path = tmp_path / "dst"
    sh.unarchive(unsafe_archive_file, dst_path, trusted=True)

    assert not dst_path.exists()
    assert unsafe_dest.exists()
    assert is_same_dir(src_dir.path, unsafe_dest / "src")


def test_unarchive__raises_when_untrusted_archive_would_extract_outside_target(
    tmp_path: Path, rep_ext: str
):
    src_dir = create_archive_source(tmp_path, File("1.txt", text="1"))

    unsafe_archive_file = tmp_path / f"unsafe{rep_ext}"
    unsafe_dest = tmp_path / "unsafe"
    create_unsafe_archive(unsafe_archive_file, src_dir.path, unsafe_dest)

    dst_path = tmp_path / "dst"
    with pytest.raises(sh.UnsafeArchiveError) as exc_info:
        sh.unarchive(unsafe_archive_file, dst_path)

    assert "destination is outside the target directory" in str(exc_info.value)
    assert not dst_path.exists()
    assert not unsafe_dest.exists()



from pathlib import Path
import stat
import typing as t

import pytest
from pytest import param

import shelmet as sh

from .utils import Dir, File


parametrize = pytest.mark.parametrize


@pytest.fixture()
def test_file(tmp_path: Path) -> Path:
    test_file = tmp_path / "test_file.txt"
    test_file.touch()
    return test_file


@pytest.fixture()
def test_dir(tmp_path: Path) -> Path:
    test_dir = tmp_path / "test_dir"
    test_dir.mkdir()
    return test_dir


@parametrize(
    "starting_mode, desired_mode, expected_mode",
    [
        param(None, 0o777, "-rwxrwxrwx"),
        param(None, 0o666, "-rw-rw-rw-"),
        param(None, "777", "-rwxrwxrwx"),
        param(None, "666", "-rw-rw-rw-"),
        param(None, "+rwx", "-rwxrwxrwx"),
        param(None, "u+rw,go-rwx", "-rw-------"),
        param(None, "=rw", "-rw-rw-rw-"),
        param(None, "a=rw", "-rw-rw-rw-"),
        param(None, "ug=rw,o=r", "-rw-rw-r--"),
        param(None, "u=rwx,g=rw,o=r", "-rwxrw-r--"),
        param(None, "=rw,o+t", "-rw-rw-rwT"),
        param(None, "=rw,a+t", "-rw-rw-rwT"),
        param(None, "=rw,+t", "-rw-rw-rwT"),
        param(None, "=rw,+x,o+t", "-rwxrwxrwt"),
        param(None, "=rw,+x,a+t", "-rwxrwxrwt"),
        param(None, "=rw,+x,+t", "-rwxrwxrwt"),
        param(None, "+rw,u+s", "-rwSrw-rw-"),
        param(None, "+rw,u+x,u+s", "-rwsrw-rw-"),
        param(None, "+rw,g+s", "-rw-rwSrw-"),
        param(None, "+rw,g+x,g+s", "-rw-rwsrw-"),
        param(0o777, "u=rwx,g=rw,o=r", "-rwxrw-r--"),
        param(0, "u=rwx,g=rw,o=r", "-rwxrw-r--"),
        param(0o600, "g=u", "-rw-rw----"),
        param(0o740, "g+u", "-rwxrwx---"),
        param(0o700, "o=u", "-rwx---rwx"),
        param(0o604, "o+u", "-rw----rw-"),
        param(0o060, "u=g", "-rw-rw----"),
        param(0o470, "u+g", "-rwxrwx---"),
        param(0o070, "o=g", "----rwxrwx"),
        param(0o064, "o+g", "----rw-rw-"),
        param(0o006, "u=o", "-rw----rw-"),
        param(0o407, "u+o", "-rwx---rwx"),
        param(0o007, "g=o", "----rwxrwx"),
        param(0o046, "g+o", "----rw-rw-"),
        param(0o600, "go=u", "-rw-rw-rw-"),
        param(0o060, "uo=g", "-rw-rw-rw-"),
        param(0o006, "ug=o", "-rw-rw-rw-"),
        param(0o600, "a=u", "-rw-rw-rw-"),
        param(0o060, "a=g", "-rw-rw-rw-"),
        param(0o006, "a=o", "-rw-rw-rw-"),
        param(0o777, "o=", "-rwxrwx---"),
    ],
)
def test_chmod__sets_file_mode(
    test_file: Path,
    starting_mode: t.Optional[int],
    desired_mode: t.Union[int, str],
    expected_mode: str,
):
    if starting_mode is not None:
        test_file.chmod(starting_mode)

    sh.chmod(test_file, desired_mode)

    filemode = stat.filemode(test_file.stat().st_mode)
    assert filemode == expected_mode


def test_chmod__sets_dir_mode(test_dir: Path):
    sh.chmod(test_dir, "+rw")

    filemode = stat.filemode(test_dir.stat().st_mode)
    assert filemode == "drwxrwxrwx"


def test_chmod__accepts_fileno(test_file: Path):
    with test_file.open() as fp:
        fd = fp.fileno()
        sh.chmod(fd, "+rwx")

    filemode = stat.filemode(test_file.stat().st_mode)
    assert filemode == "-rwxrwxrwx"


@parametrize(
    "items, mode, expected_file_mode, expected_dir_mode",
    [
        param(
            [
                Dir("a", File("1.txt"), File("2.txt"), File("3.txt")),
                Dir(
                    "b",
                    Dir("c", File("4.txt"), Dir("d", File("5.txt"))),
                    File("6.txt"),
                    File("7.txt"),
                    File("8.txt"),
                ),
                File("9.txt"),
                File("10.txt"),
            ],
            "766",
            "-rwxrw-rw-",
            "drwxrw-rw-",
        ),
        param(
            [
                Dir("a", File("1.txt"), File("2.txt"), File("3.txt")),
                Dir(
                    "b",
                    Dir("c", File("4.txt"), Dir("d", File("5.txt"))),
                    File("6.txt"),
                    File("7.txt"),
                    File("8.txt"),
                ),
                File("9.txt"),
                File("10.txt"),
            ],
            "go-rwx",
            "-rw-------",
            "drwx------",
        ),
    ],
)
def test_chmod__recursively_sets_mode(
    tmp_path: Path,
    items: t.List[t.Union[Dir, File]],
    mode: t.Union[int, str],
    expected_file_mode: int,
    expected_dir_mode: int,
):
    test_dir = Dir(tmp_path / "test_dir", *items)
    test_dir.mkdir()

    sh.chmod(test_dir.path, mode, recursive=True)

    for path in (test_dir.path, *sh.walk(test_dir.path)):
        expected_mode = expected_dir_mode if path.is_dir() else expected_file_mode
        path_mode = stat.filemode(path.stat().st_mode)

        assert (
            path_mode == expected_mode
        ), f"Expected mode of {path} to be {expected_mode!r}, not {path_mode!r}"


@parametrize(
    "mode, exception",
    [
        param(None, TypeError),
        param("", ValueError),
        param("u=Z", ValueError),
        param("u=rwxg", ValueError),
        param("rw", ValueError),
        param("u=t", ValueError),
    ],
)
def test_chmod__raises_when_mode_invalid(
    test_file: Path, mode: t.Any, exception: t.Type[Exception]
):
    with pytest.raises(exception):
        sh.chmod(test_file, mode)



"""The archiving module contains utilities for interacting with archive files."""

from abc import ABC, abstractmethod
from datetime import datetime, timezone
import errno
import os
from pathlib import Path, PurePath
import tarfile
from types import TracebackType
import typing as t
import zipfile

from .fileio import atomicfile
from .filesystem import cp
from .path import Ls, walk
from .types import StrPath


try:
    import zlib
except ImportError:  # pragma: no cover
    zlib = None  # type: ignore


# Use same default tar format for older Python versions for consistency (default was changed to PAX
# in 3.8).
# NOTE: This format is only used for writing and doesn't affect reading archives in other formats.
DEFAULT_TAR_FORMAT = tarfile.PAX_FORMAT

# Use ZIP_DEFLATED as default zipfile compression if available.
DEFAULT_ZIP_COMPRESSION = zipfile.ZIP_DEFLATED if zlib else zipfile.ZIP_STORED

# Archive names to exclude when adding to an archive.
EXCLUDE_ARCNAMES = {".", ".."}


class ArchiveError(Exception):
    """General archive error."""

    def __init__(self, *args: t.Any, orig_exc: t.Optional[Exception] = None):
        super().__init__(*args)
        self.orig_exc = orig_exc


class UnsafeArchiveError(ArchiveError):
    """Unsafe archive exception raised when an untrusted archive would extract contents outside of
    the destination directory."""

    pass


class ArchiveSource:
    """Iterable representation of a path that should be added to an archive."""

    def __init__(self, source: t.Union[StrPath, Ls]):
        path = Path(source).resolve()
        subpaths: t.Optional[Ls] = None

        if isinstance(source, Ls):
            subpaths = source
        elif path.is_dir():
            subpaths = walk(source)

        self.source = source
        self.path = path
        self.subpaths = subpaths

    def __repr__(self) -> str:
        if isinstance(self.source, Ls):
            source = repr(self.source)
        else:
            source = f"'{self.source}'"
        return f"{self.__class__.__name__}(source={source}, path='{self.path}')"

    def __str__(self) -> str:
        """Return string representation of archive source."""
        return str(self.source)

    def __iter__(self) -> t.Iterator[Path]:
        """Yield contents of archive source including the base path and its subpaths."""
        yield self.path

        if not self.subpaths:
            return

        for subpath in self.subpaths:
            yield Path(subpath).resolve()


class BaseArchive(ABC):
    """Abstract base class that provides a common interface for interacting with different archive
    formats."""

    def __init__(self, backend):
        self.backend = backend

    def __enter__(self) -> "BaseArchive":
        """Enter context manager when reading or writing an archive."""
        return self

    def __exit__(
        self,
        exc_type: t.Optional[t.Type[BaseException]],
        exc_val: t.Optional[BaseException],
        exc_tb: t.Optional[TracebackType],
    ) -> None:
        """Exit context manager after reading or writing an archive."""
        self.close()

    @classmethod
    @abstractmethod
    def open(cls, file: t.Union[StrPath, t.IO], mode: str = "r") -> "BaseArchive":
        """Open an archive file."""
        pass  # pragma: no cover

    @abstractmethod
    def close(self) -> None:
        """Close the archive file."""
        pass  # pragma: no cover

    @abstractmethod
    def list(self) -> t.List[str]:
        """Return a list of file/directory names contained in the archive."""
        pass  # pragma: no cover

    @abstractmethod
    def extractall(self, path: StrPath) -> None:
        """Extract all contents of the archive to the given path."""
        pass  # pragma: no cover

    @abstractmethod
    def add(self, path: StrPath, arcname: t.Optional[str] = None) -> None:
        """Add path to the archive non-recursively."""
        pass  # pragma: no cover

    def addsource(self, source: ArchiveSource, arcname: t.Optional[StrPath] = None) -> None:
        """Add file system contents of source to archive."""
        if arcname:
            arcname = Path(arcname)

        root_path_offset = len(source.path.parts)

        for path in source:
            if arcname:
                name = str(Path(arcname, *path.parts[root_path_offset:]))
            else:  # pragma: no cover
                name = str(path)

            if name in EXCLUDE_ARCNAMES:
                continue

            self.add(path, arcname=name)


class ZipArchive(BaseArchive):
    """Archive class for interacting with zip archives."""

    backend: zipfile.ZipFile

    @classmethod
    def open(cls, file: t.Union[StrPath, t.IO], mode: str = "r") -> "ZipArchive":
        """Open an archive file."""
        return cls(zipfile.ZipFile(file, mode, compression=DEFAULT_ZIP_COMPRESSION))

    def close(self):
        """Close the archive file."""
        self.backend.close()

    def list(self) -> t.List[str]:
        """Return a list of file/directory names contained in the archive."""
        return self.backend.namelist()

    def extractall(self, path: StrPath) -> None:
        """Extract all contents of the archive to the given path."""
        self.backend.extractall(path)

    def add(self, path: StrPath, arcname: t.Optional[str] = None) -> None:
        """Add path to the archive non-recursively."""
        if not isinstance(path, Path):  # pragma: no cover
            path = Path(path)

        if not arcname:  # pragma: no cover
            arcname = str(path)

        self.backend.write(path, arcname=arcname)


class TarArchive(BaseArchive):
    """Archive class for interacting with tar archives without compression."""

    backend: tarfile.TarFile
    compression = ""

    @classmethod
    def open(cls, file: t.Union[StrPath, t.IO], mode: str = "r") -> "TarArchive":
        """Open an archive file."""
        if mode == "w" and cls.compression:
            mode = f"{mode}:{cls.compression}"

        if isinstance(file, (str, bytes, Path)):
            tar = tarfile.open(file, mode=mode, format=DEFAULT_TAR_FORMAT)
        else:
            tar = tarfile.open(fileobj=file, mode=mode, format=DEFAULT_TAR_FORMAT)

        return cls(tar)

    def close(self):
        """Close the archive file."""
        self.backend.close()

    def list(self) -> t.List[str]:
        """Return a list of file/directory names contained in the archive."""
        return self.backend.getnames()

    def extractall(self, path: StrPath) -> None:
        """Extract all contents of the archive to the given path."""
        self.backend.extractall(path)

    def add(self, path: StrPath, arcname: t.Optional[str] = None) -> None:
        """Add path to the archive non-recursively."""
        self.backend.add(path, arcname=arcname, recursive=False)


class TarGzArchive(TarArchive):
    """Archive class for interacting with tar archives with gzip compression."""

    compression = "gz"


class TarBzArchive(TarArchive):
    """Archive class for interacting with tar archives with bzip2 compression."""

    compression = "bz2"


class TarXzArchive(TarArchive):
    """Archive class for interacting with tar archives with lzma compression."""

    compression = "xz"


EXTENSION_ARCHIVES: t.Dict[str, t.Type[BaseArchive]] = {
    # Extensions that use uncompressed tar.
    ".tar": TarArchive,
    # Extensions that use Tar+gz compression.
    ".tar.gz": TarGzArchive,
    ".tgz": TarGzArchive,
    ".taz": TarGzArchive,
    # Extensions that use tar+bz2 compression.
    ".tar.bz2": TarBzArchive,
    ".tb2": TarBzArchive,
    ".tbz": TarBzArchive,
    ".tbz2": TarBzArchive,
    ".tz2": TarBzArchive,
    # Extensions that use tar+xz compression.
    ".tar.xz": TarXzArchive,
    ".txz": TarXzArchive,
    # Extensions that use zip format.
    ".docx": ZipArchive,
    ".egg": ZipArchive,
    ".jar": ZipArchive,
    ".odg": ZipArchive,
    ".odp": ZipArchive,
    ".ods": ZipArchive,
    ".odt": ZipArchive,
    ".pptx": ZipArchive,
    ".xlsx": ZipArchive,
    ".zip": ZipArchive,
}


def archive(
    file: StrPath,
    *paths: t.Union[StrPath, Ls],
    root: t.Optional[StrPath] = None,
    repath: t.Optional[t.Union[str, t.Mapping[StrPath, StrPath]]] = None,
    ext: str = "",
) -> None:
    """
    Create an archive from the given source paths.

    The source paths can be relative or absolute but the path names inside the archive will always
    be relative. By default, the paths within the archive will be determined by taking the common
    path of all the sources and removing it from each source path so that the archive paths are all
    relative to the shared parent path of all sources. If `root` is given, it will be used in place
    of the dynamic common path determination, but it must be a parent path common to all sources.

    The archive member names of the source paths can be customized using the `repath` argument. The
    `repath` argument is a mapping of source paths to their custom archive name. If a source path is
    given as relative, then its repath key must also be relative. If a source path is given as
    absolute, then its repath key must also be absolute. The repath keys/values should be either
    strings or ``Path`` objects but they don't have to match the corresponding source path. Both the
    keys and values will have their path separators normalized.

    Archives can be created in either the tar or zip format. A tar archive can use the same
    compressions that are available from ``tarfile`` which are gzipped, bzip2, and lzma. A zip
    archive will use deflate compression if the ``zlib`` library is available. Otherwise, it will
    fallback to being uncompressed.

    The archive format is interfered from the file extension of `file` by default, but can be
    overridden using the `ext` argument (e.g. ``ext=".tgz"`` for a gzipped tarball).

    The supported tar-based extensions are:

    - ``.tar``
    - ``.tar.gz``, ``.tgz``, ``.taz``
    - ``.tar.bz2``, ``.tb2``, ``.tbz``, ``.tbz2``, ``.tz2``
    - ``.tar.xz``, ``.txz``

    The supported zip-based extensions are:

    - ``.zip``,
    - ``.egg``, ``.jar``
    - ``.docx``, ``pptx``, ``xlsx``
    - ``.odg``, ``.odp``, ``.ods``, ``.odt``

    Args:
        file: Archive file path to create.
        *paths: Source paths (files and/or directories) to archive. Directories will be recursively
            added.
        root: Archive member paths will be relative to this root directory. The root path must be a
            parent directory of all source paths, otherwise, an exception will be raised.
        repath: A mapping of source paths to archive names that will rename the source path to the
            mapped value within the archive. A string representing the archive member name can only
            be used when a single source path is being added to the archive.
        ext: Specify the archive format to use by referencing the corresponding file extension
            (starting with a leading ".") instead of interfering the format from the `file`
            extension.
    """
    file = Path(file)
    archive_class = _get_archive_class_or_raise(file, ext)

    if repath is None:
        repath = {}

    if isinstance(repath, str) and len(paths) > 1:
        raise TypeError("repath must be a dict when there is more than one archive source path")

    if not isinstance(repath, str) and not isinstance(repath, dict):
        raise TypeError("repath must be a string or dict")

    sources = [ArchiveSource(path) for path in paths]

    if isinstance(repath, str):
        repath = {str(sources[0]): Path(repath)}
    else:
        repath = {str(Path(src)): Path(pth) for src, pth in repath.items()}

    if root:
        root = Path(root).resolve()
    else:
        # The archive contents will be relative to the common path shared by all source paths.
        root = Path(os.path.commonpath([src.path for src in sources])).parent

    # Check that source paths are valid relative to root before adding archive members. No need to
    # check sources that are going to be repathed since their arcname won't depend on them being
    # relative to the root directory.
    _verify_archive_root(root, [source for source in sources if str(source) not in repath])

    # Use atomicfile so that archive is only created at location if there are no errors while
    # archiving all paths.
    with atomicfile(file, "wb", skip_sync=True) as fp:
        with archive_class.open(fp, "w") as archive_file:
            try:
                for source in sources:
                    arcname = repath.get(str(source))
                    if arcname is None:
                        arcname = str(source.path.relative_to(root))
                    archive_file.addsource(source, arcname=arcname)
            except Exception as exc:
                raise ArchiveError(
                    f"archive: Failed to create archive '{file}' due to error: {exc}", orig_exc=exc
                ) from exc


def backup(
    src: StrPath,
    *,
    timestamp: t.Optional[str] = "%Y-%m-%dT%H:%M:%S.%f%z",
    utc: bool = False,
    epoch: bool = False,
    prefix: str = "",
    suffix: str = "~",
    ext: t.Optional[str] = None,
    hidden: bool = False,
    overwrite: bool = False,
    dir: t.Optional[StrPath] = None,
    namer: t.Optional[t.Callable[[Path], StrPath]] = None,
) -> Path:
    """
    Create a backup of a file or directory as either a direct copy or an archive file.

    The format of the backup name is ``{prefix}{src}.{timestamp}{suffix|ext}``.

    By default, the backup will be created in the same parent directory as the source and be named
    like ``"src.YYYY-MM-DDThh:mm:ss.ffffff~"``, where the timestamp is the current local time.

    If `utc` is ``True``, then the timestamp will be in the UTC timezone.

    If `epoch` is ``True``, then the timestamp will be the Unix time as returned by
    ``time.time()`` instead of the strftime format.

    If `ext` is given, the backup created will be an archive file. The extension must be one that
    :func:`archive` supports. The `suffix` value will be ignored and `ext` used in its place.

    If `hidden` is ``True``, then a ``"."`` will be prepended to the `prefix`. It won't be added if
    `prefix` already starts with a ``"."``.

    If `dir` is given, it will be used as the parent directory of the backup instead of the source's
    parent directory.

    If `overwrite` is ``True`` and the backup location already exists, then it will be overwritten.

    If `namer` is given, it will be called with ``namer(src)`` and it should return the full
    destination path of the backup. All other arguments to this function will be ignored except for
    `overwrite`.

    Args:
        src: Source file or directory to backup.
        timestamp: Timestamp strftime-format string or ``None`` to exclude timestamp from backup
            name. Defaults to ISO-8601 format.
        utc: Whether to use UTC time instead of local time for the timestamp.
        epoch: Whether to use the Unix time for the timestamp instead of the strftime format in
            `timestamp`.
        prefix: Name prefix to prepend to the backup.
        suffix: Name suffix to append to the backup.
        ext: Create an archive of `src` as the backup instead of a direct copy using the given
            archive extension. The extension must be supported by :func:`archive` or an exception
            will be raised. When given the `suffix` value is ignored and `ext` will be used in its
            place.
        hidden: Whether to ensure that the backup location is a hidden file or directory.
        overwrite: Whether to overwrite an existing file or directory when backing up.
        dir: Set the parent directory of the backup. Defaults to ``None`` which will use the parent
            directory of the `src`.
        namer: Naming function that can be used to return the full path of the backup location. It
            will be passed the `src` value as a ``pathlib.Path`` object as a positional argument. It
            should return the destination path of the backup as a ``str`` or ``pathlib.Path``.

    Returns:
        Backup location.
    """
    if not isinstance(timestamp, str) and timestamp is not None:
        raise ValueError(
            f"timestamp should be a strftime-formatted string or None, not {timestamp!r}"
        )

    src = Path(src).resolve()

    if ext:
        suffix = ext

    if namer:
        dst = Path(namer(src)).resolve()
    else:
        dst = _backup_namer(
            src,
            timestamp=timestamp,
            utc=utc,
            epoch=epoch,
            prefix=prefix,
            suffix=suffix,
            hidden=hidden,
            dir=dir,
        )

    if src == dst:
        raise FileExistsError(errno.EEXIST, f"Backup destination cannot be the source: {src}")

    if not overwrite and dst.exists():
        raise FileExistsError(errno.EEXIST, f"Backup destination already exists: {dst}")

    if ext:
        archive(dst, src, ext=ext)
    else:
        cp(src, dst)

    return dst


def _backup_namer(
    src: Path,
    *,
    timestamp: t.Optional[str] = "%Y-%m-%dT%H:%M:%S.%f%z",
    utc: bool = False,
    epoch: bool = False,
    prefix: str = "",
    suffix: str = "~",
    hidden: bool = False,
    dir: t.Optional[StrPath] = None,
) -> Path:
    if not dir:
        dir = src.parent.resolve()
    else:
        dir = Path(dir).resolve()

    if hidden and not prefix.startswith("."):
        prefix = f".{prefix}"

    ts: t.Union[str, float] = ""
    if timestamp is not None:
        tz = None
        if utc:
            tz = timezone.utc
        dt = datetime.now(tz)

        if epoch:
            ts = dt.timestamp()
        else:
            ts = dt.strftime(timestamp)

        ts = f".{ts}"

    name = f"{prefix}{src.name}{ts}{suffix}"
    dst = dir / name

    return dst


def lsarchive(file: StrPath, ext: str = "") -> t.List[PurePath]:
    """
    Return list of member paths contained in archive file.

    Args:
        file: Archive file to list.
        ext: Specify the archive format to use by referencing the corresponding file extension
            (starting with a leading ".") instead of interfering the format from the `file`
            extension.
    """
    file = Path(file)
    archive_class = _get_archive_class_or_raise(file, ext)
    with archive_class.open(file, "r") as archive_file:
        return [PurePath(item) for item in archive_file.list()]


def unarchive(file: StrPath, dst: StrPath = ".", *, ext: str = "", trusted: bool = False) -> None:
    """
    Extract an archive to the given destination path.

    If the archive contains any paths that would be extracted outside the destination path, an
    :class:`ArchiveError` will be raised to prevent untrusted archives from extracting contents to
    locations that may pose a security risk. To allow a trusted archive to extract contents outside
    the destination, use the argument ``trusted=True``.

    Archives can be extracted from either zip or tar formats with compression. The tar compressions
    available are the same as what is supported by ``tarfile`` which are gzipped, bzip2, and lzma.

    The archive format is interfered from the file extension of `file` by default, but can be
    overridden using the `ext` argument (e.g. ``ext=".tgz"`` for a gzipped tarball).

    The supported tar extensions are:

    - ``.tar``
    - ``.tar.gz``, ``.tgz``, ``.taz``
    - ``.tar.bz2``, ``.tb2``, ``.tbz``, ``.tbz2``, ``.tz2``
    - ``.tar.xz``, ``.txz``
    - ``.zip``,
    - ``.egg``, ``.jar``
    - ``.docx``, ``pptx``, ``xlsx``
    - ``.odg``, ``.odp``, ``.ods``, ``.odt``

    Args:
        file: Archive file to unarchive.
        dst: Destination directory to unarchive contents to.
        ext: Specify the archive format to use by referencing the corresponding file extension
            (starting with a leading ".") instead of interfering the format from the `file`
            extension.
        trusted: Whether the archive is safe and can be trusted to allow it to extract contents
            outside of the destination path. Only enable this for archives that have been verified
            as originating from a trusted source.
    """
    file = Path(file)
    archive_class = _get_archive_class_or_raise(file, ext)

    try:
        with archive_class.open(file, "r") as archive_file:
            if not trusted:
                _verify_archive_safety(archive_file, dst)
            archive_file.extractall(dst)
    except ArchiveError:  # pragma: no cover
        raise
    except Exception as exc:
        raise ArchiveError(
            f"unarchive: Failed to unarchive '{file}' due to error: {exc}", orig_exc=exc
        ) from exc


def _get_archive_class_or_raise(file: Path, ext: str = "") -> t.Type[BaseArchive]:
    """Return the :class:`BaseArchive` that should be used to handle an archive file or raise if
    none found."""
    archive_class = _get_archive_class(file, ext=ext)
    if not archive_class:
        raise NotImplementedError(f"Archive format not supported: {ext!r}")
    return archive_class


def _get_archive_class(file: Path, ext: str = "") -> t.Optional[t.Type[BaseArchive]]:
    """Return the :class:`BaseArchive` that should be used to handle an archive file."""
    archive_class = EXTENSION_ARCHIVES.get(ext)

    if not archive_class:
        archive_class = EXTENSION_ARCHIVES.get("".join(file.suffixes))

    if not archive_class:
        archive_class = next(
            (a for e, a in EXTENSION_ARCHIVES.items() if file.name.endswith(e)), None
        )

    return archive_class


def _verify_archive_root(root: Path, sources: t.List[ArchiveSource]) -> None:
    """Check whether archive root path is valid for sources before adding them to an archive."""
    for source in sources:
        try:
            source.path.relative_to(root)
        except ValueError:
            raise ValueError(
                f"Source paths must be a subpath of the root archive path. '{source.path}' is"
                f" not in the subpath of '{root}'"
            )


def _verify_archive_safety(archive_file: BaseArchive, dst: StrPath) -> None:
    """Check whether the archive contains paths that would be extracted outside the target path and
    raise an exception if it would."""
    dst = Path(dst).resolve()
    safe_path_prefix = str(dst)

    for name in archive_file.list():
        extraction_path = (dst / name).resolve()
        if not str(extraction_path).startswith(safe_path_prefix):
            raise UnsafeArchiveError(
                f"unarchive: Archive has member '{name}' whose destination is outside the"
                f" target directory '{dst}' and cannot be extracted unless it is designated as"
                f" originating from a trusted source with 'trusted=True`."
            )



from pathlib import Path

import shelmet as sh


def test_umask(tmp_path: Path):
    mode = 0o644  # -rw-rw-r-- (user,group = read-write, other = read)
    umask = 0o77  # g-rw,o-rw (disallow read-write for group and other)
    expected_mode_with_umask = 0o600  # -rw------- (user = read-write, group,other = no-access)

    file_before_umask = tmp_path / "before"
    file_with_umask = tmp_path / "with"
    file_after_umask = tmp_path / "after"

    file_before_umask.touch(mode=mode)
    stat_before = file_before_umask.stat()
    assert oct(stat_before.st_mode & 0o777) == oct(mode)

    with sh.umask(umask):
        file_with_umask.touch(mode=mode)
        stat_during = file_with_umask.stat()
        assert oct(stat_during.st_mode & 0o777) == oct(expected_mode_with_umask)

    file_after_umask.touch(mode=mode)
    stat_after = file_after_umask.stat()
    assert oct(stat_after.st_mode & 0o777) == oct(mode)



from pathlib import Path
import typing as t

import pytest
from pytest import param

import shelmet as sh


parametrize = pytest.mark.parametrize


@parametrize(
    "paths,",
    [
        param(["a"]),
        param(["a", "a/b", "a/b/c", "d/e/f/g/h"]),
        param([Path("a")]),
        param([Path("a"), Path("a/b"), Path("a/b/c"), Path("d/e/f/g/h")]),
    ],
)
def test_mkdir(tmp_path: Path, paths: t.List[t.Union[str, Path]]):
    targets = [tmp_path / path for path in paths]
    sh.mkdir(*targets)

    for target in targets:
        assert target.is_dir()


def test_mkdir__sets_mode(tmp_path: Path):
    targets = [tmp_path / "test1", tmp_path / "test2", tmp_path / "1" / "2" / "3"]
    mode = 0o755

    sh.mkdir(*targets, mode=mode)
    for target in targets:
        target_mode = oct(target.stat().st_mode & 0o777)
        assert target_mode == oct(mode)


def test_mkdir__raises_if_exist_not_ok(tmp_path: Path):
    with pytest.raises(FileExistsError):
        sh.mkdir(tmp_path, exist_ok=False)



"""
This module provides the CLI interface for invoke tasks.

All tasks can be executed from this file's directory using:

    $ inv <task>

Where <task> is a function defined below with the @task decorator.
"""

from functools import partial
import os

from invoke import Exit, UnexpectedExit, run as _run, task


PACKAGE_NAME = "shelmet"
PACKAGE_SOURCE = f"src/{PACKAGE_NAME}"
TEST_TARGETS = f"{PACKAGE_SOURCE} tests"
LINT_TARGETS = f"{TEST_TARGETS} tasks.py"
EXIT_EXCEPTIONS = (Exit, UnexpectedExit, SystemExit)


# Set pyt=True to enable colored output when available.
run = partial(_run, pty=True)


@task
def black(ctx, quiet=False):
    """Autoformat code using black."""
    run(f"black {LINT_TARGETS}", hide=quiet)


@task
def isort(ctx, quiet=False):
    """Autoformat Python imports."""
    run(f"isort {LINT_TARGETS}", hide=quiet)


@task
def docformatter(ctx):
    """Autoformat docstrings using docformatter."""
    run(
        f"docformatter -r {LINT_TARGETS} "
        f"--in-place --pre-summary-newline --wrap-descriptions 100 --wrap-summaries 100"
    )


@task
def fmt(ctx):
    """Autoformat code and docstrings."""
    print("Running docformatter")
    docformatter(ctx)

    print("Running isort")
    isort(ctx, quiet=True)

    print("Running black")
    black(ctx, quiet=True)


@task
def flake8(ctx):
    """Check code for PEP8 violations using flake8."""
    run(f"flake8 --format=pylint {LINT_TARGETS}")


@task
def pylint(ctx):
    """Check code for static errors using pylint."""
    run(f"pylint {LINT_TARGETS}")


@task
def mypy(ctx):
    """Check code using mypy type checker."""
    run(f"mypy {LINT_TARGETS}")


@task
def lint(ctx):
    """Run linters."""
    linters = {"flake8": flake8, "pylint": pylint, "mypy": mypy}
    failures = []

    for name, linter in linters.items():
        print(f"Running {name}")
        try:
            linter(ctx)
        except EXIT_EXCEPTIONS:
            failures.append(name)
            result = "FAILED"
        else:
            result = "PASSED"
        print(f"{result}\n")

    if failures:
        failed = ", ".join(failures)
        raise Exit(f"ERROR: Linters that failed: {failed}")


@task(help={"args": "Override default pytest arguments"})
def unit(ctx, args=f"{TEST_TARGETS} --cov={PACKAGE_NAME} --flake8 --mypy --pylint"):
    """Run unit tests using pytest."""
    tox_env_site_packages_dir = os.getenv("TOX_ENV_SITE_PACKAGES_DIR")
    if tox_env_site_packages_dir:
        # Re-path package source to match tox env so that we generate proper coverage report.
        tox_env_pkg_src = os.path.join(tox_env_site_packages_dir, os.path.basename(PACKAGE_SOURCE))
        args = args.replace(PACKAGE_SOURCE, tox_env_pkg_src)

    run(f"pytest {args}")


@task
def test(ctx):
    """Run linters and tests."""
    print("Building package")
    build(ctx)

    print("Building docs")
    docs(ctx)

    print("Checking linters")
    lint(ctx)

    print("Running unit tests")
    unit(ctx, args=f"{TEST_TARGETS} --cov={PACKAGE_NAME}")


@task
def docs(ctx, serve=False, bind="127.0.0.1", port=8000):
    """Build docs."""
    run("rm -rf docs/_build")
    run("sphinx-build -q -W -b html docs docs/_build/html")

    if serve:
        print(f"Serving docs on {bind} port {port} (http://{bind}:{port}/) ...")
        run(f"python -m http.server -b {bind} --directory docs/_build/html {port}", hide=True)


@task
def build(ctx):
    """Build Python package."""
    run("rm -rf dist build docs/_build")
    run("python setup.py -q sdist bdist_wheel")


@task
def clean(ctx):
    """Remove temporary files related to development."""
    run("find . -type f -name '*.py[cod]' -delete -o -type d -name __pycache__ -delete")
    run("rm -rf .tox .coverage .cache .pytest_cache **/.egg* **/*.egg* dist build")


@task(pre=[build])
def release(ctx):
    """Release Python package."""
    run("twine upload dist/*")



from pathlib import Path
import typing as t

import pytest
from pytest import param

import shelmet as sh


parametrize = pytest.mark.parametrize


@parametrize(
    "paths, expected",
    [
        param(["a"], "a"),
        param(["a/"], "a"),
        param(["a", "b", "c/d"], "a/b/c/d"),
        param(["a", "/b", "/c/d"], "a/b/c/d"),
        param(["/a", "b", "c/d"], "/a/b/c/d"),
        param(["/a/", "/b/", "/c/d/"], "/a/b/c/d"),
        param([Path("a")], "a"),
        param([Path("a/")], "a"),
        param([Path("a"), Path("b"), Path("c/d")], "a/b/c/d"),
        param([Path("a"), Path("/b"), Path("/c/d")], "a/b/c/d"),
        param([Path("/a"), Path("b"), Path("c/d")], "/a/b/c/d"),
        param(["a", Path("b"), "c/d"], "a/b/c/d"),
        param([Path("a"), "b", Path("c/d")], "a/b/c/d"),
    ],
)
def test_reljoin(paths: t.Sequence[t.Union[Path, str]], expected: str):
    assert sh.reljoin(*paths) == expected
